"""Core definitions to be shared amongst the modules."""
# standard library imports
import io
import logging
import shutil
import struct
from typing import Tuple
from uuid import UUID

# local imports
from . import jp2box
from .lib.tiff import DATATYPE2FMT
from .jp2k import Jp2k
from glymur.core import RESTRICTED_ICC_PROFILE

# Mnemonics for the two TIFF format version numbers.
TIFF = 42
BIGTIFF = 43


class _2JP2Converter(object):
    """
    This private class is used by both the TIFF2JP2 and the JPEG2JP2
    converters.

    Attributes
    ----------
    create_exif_uuid : bool
        Create a UUIDBox for the TIFF IFD metadata.
    tilesize : tuple
        The dimensions of a tile in the JP2K file.
    verbosity : int
        Set the level of logging, i.e. WARNING, INFO, etc.
    """

    def __init__(
        self,
        create_exif_uuid: bool,
        create_xmp_uuid: bool,
        include_icc_profile: bool,
        tilesize: Tuple[int, int] | None,
        verbosity: int
    ):

        self.create_exif_uuid = create_exif_uuid
        self.create_xmp_uuid = create_xmp_uuid
        self.include_icc_profile = include_icc_profile
        self.tilesize = tilesize

        # Assume that there is no ICC profile tag until we know otherwise.
        self.icc_profile = None

        self.setup_logging(verbosity)

    def read_ifd(self, tfp):
        """Process either the main IFD or an Exif IFD

        Parameters
        ----------
        tfp : file-like
            FILE pointer for TIFF

        Returns
        -------
        dictionary of the TIFF IFD
        """

        self.found_geotiff_tags = False

        tag_length = 20 if self.version == BIGTIFF else 12

        # how many tags?
        if self.version == BIGTIFF:
            buffer = tfp.read(8)
            (num_tags,) = struct.unpack(self.endian + "Q", buffer)
        else:
            buffer = tfp.read(2)
            (num_tags,) = struct.unpack(self.endian + "H", buffer)

        # Ok, so now we have the IFD main body, but following that we have
        # the tag payloads that cannot fit into 4 bytes.

        # the IFD main body in the TIFF.  As it might be big endian, we
        # cannot just process it as one big chunk.
        buffer = tfp.read(num_tags * tag_length)

        if self.version == BIGTIFF:
            tag_format_str = self.endian + "HHQQ"
            tag_payload_offset = 12
            max_tag_payload_length = 8
        else:
            tag_format_str = self.endian + "HHII"
            tag_payload_offset = 8
            max_tag_payload_length = 4

        tags = {}

        for idx in range(num_tags):

            self.logger.debug(f"tag #: {idx}")

            tag_data = buffer[idx * tag_length:(idx + 1) * tag_length]

            tag, dtype, nvalues, offset = struct.unpack(
                tag_format_str, tag_data
            )  # noqa : E501

            if tag == 34735:
                self.found_geotiff_tags = True

            payload_length = DATATYPE2FMT[dtype]["nbytes"] * nvalues

            if tag in (34665, 34853):

                # found exif or gps ifd
                # save our location, go get that IFD, and come on back
                orig_pos = tfp.tell()
                tfp.seek(offset)
                payload = self.read_ifd(tfp)
                tfp.seek(orig_pos)

            elif payload_length > max_tag_payload_length:
                # the payload does not fit into the tag entry, so use the
                # offset to seek to that position
                current_position = tfp.tell()
                tfp.seek(offset)
                payload_buffer = tfp.read(payload_length)
                tfp.seek(current_position)

                # read the payload from the TIFF
                payload_format = DATATYPE2FMT[dtype]["format"] * nvalues
                payload = struct.unpack(
                    self.endian + payload_format,
                    payload_buffer
                )

            else:
                # the payload DOES fit into the TIFF tag entry
                payload_buffer = tag_data[tag_payload_offset:]

                # read ALL of the payload buffer
                fmt = DATATYPE2FMT[dtype]["format"]
                nelts = max_tag_payload_length / DATATYPE2FMT[dtype]["nbytes"]
                num_items = int(nelts)
                payload_format = self.endian + fmt * num_items
                payload = struct.unpack(payload_format, payload_buffer)

                # Extract the actual payload.  Two things going
                # on here.  First of all, not all of the items may
                # be used.  For example, if the payload length is
                # 4 bytes but the format string was HHH, the that
                # last 16 bit value is not wanted, so we should
                # discard it.  Second thing is that the signed and
                # unsigned rational datatypes effectively have twice
                # the number of values so we need to account for that.
                if dtype in [5, 10]:
                    payload = payload[: 2 * nvalues]
                else:
                    payload = payload[:nvalues]

            tags[tag] = {"dtype": dtype, "nvalues": nvalues, "payload": payload}

        return tags

    def setup_logging(self, verbosity):

        self.logger = logging.getLogger("tiff2jp2")
        self.logger.setLevel(verbosity)
        ch = logging.StreamHandler()
        ch.setLevel(verbosity)
        self.logger.addHandler(ch)

    def read_tiff_header(self, tfp):
        """Get the endian-ness of the TIFF, seek to the main IFD"""

        buffer = tfp.read(4)
        data = struct.unpack("BB", buffer[:2])

        # big endian or little endian?
        if data[0] == 73 and data[1] == 73:
            # little endian
            self.endian = "<"
        elif data[0] == 77 and data[1] == 77:
            # big endian
            self.endian = ">"
        # no other option is possible, libtiff.open would have errored out
        # else:
        #     msg = (
        #         f"The byte order indication in the TIFF header "
        #         f"({data}) is invalid.  It should be either "
        #         f"{bytes([73, 73])} or {bytes([77, 77])}."
        #     )
        #     raise RuntimeError(msg)

        # version number and offset to the first IFD
        (version,) = struct.unpack(self.endian + "H", buffer[2:4])
        self.version = TIFF if version == 42 else BIGTIFF

        if self.version == BIGTIFF:
            buffer = tfp.read(12)
            _, _, offset = struct.unpack(self.endian + "HHQ", buffer)
        else:
            buffer = tfp.read(4)
            (offset,) = struct.unpack(self.endian + "I", buffer)
        tfp.seek(offset)

    def append_exif_uuid_box(self):
        """Append an EXIF UUID box onto the end of the JPEG 2000 file.  It will
        contain metadata from the TIFF IFD.
        """
        if not self.create_exif_uuid:
            return

        # create a bytesio object for the IFD
        b = io.BytesIO()

        # write this 32-bit header into the UUID, no matter if we had bigtiff
        # or regular tiff or big endian
        data = struct.pack("<BBHI", 73, 73, 42, 8)
        b.write(data)

        self.write_ifd(b, self.tags)

        # create the Exif UUID
        if self.found_geotiff_tags:
            # geotiff UUID
            the_uuid = UUID("b14bf8bd-083d-4b43-a5ae-8cd7d5a6ce03")
            payload = b.getvalue()
        else:
            # Make it an exif UUID.
            the_uuid = UUID(bytes=b"JpgTiffExif->JP2")
            payload = b"EXIF\0\0" + b.getvalue()

        # the length of the box is the length of the payload plus 8 bytes
        # to store the length of the box and the box ID
        box_length = len(payload) + 8

        uuid_box = jp2box.UUIDBox(the_uuid, payload, box_length)
        with self.jp2_path.open(mode="ab") as f:
            uuid_box.write(f)

        self.jp2.finalize(force_parse=True)

    def write_ifd(self, b, tags):
        """Write the IFD out to the UUIDBox.  We will always write IFDs
        for 32-bit TIFFs, i.e. 12 byte tags, meaning just 4 bytes within
        the tag for the tag data
        """

        little_tiff_tag_length = 12
        max_tag_payload_length = 4

        # exclude any unwanted tags
        if self.exclude_tags is not None:
            for tag in self.exclude_tags:
                if tag in tags:
                    tags.pop(tag)

        num_tags = len(tags)
        write_buffer = struct.pack("<H", num_tags)
        b.write(write_buffer)

        # Ok, so now we have the IFD main body, but following that we have
        # the tag payloads that cannot fit into 4 bytes.

        ifd_start_loc = b.tell()
        after_ifd_position = ifd_start_loc + num_tags * little_tiff_tag_length

        for idx, tag in enumerate(tags):

            tag_offset = ifd_start_loc + idx * little_tiff_tag_length
            self.logger.debug(f"tag #: {tag}, writing to {tag_offset}")
            self.logger.debug(f"tag #: {tag}, after IFD {after_ifd_position}")

            b.seek(tag_offset)

            try:
                dtype = tags[tag]["dtype"]
            except IndexError:
                breakpoint()
                pass

            nvalues = tags[tag]["nvalues"]
            payload = tags[tag]["payload"]

            payload_length = DATATYPE2FMT[dtype]["nbytes"] * nvalues

            if payload_length > max_tag_payload_length:

                # the payload does not fit into the tag entry

                # read the payload from the TIFF
                payload_format = DATATYPE2FMT[dtype]["format"] * nvalues

                # write the tag entry to the UUID
                new_offset = after_ifd_position
                buffer = struct.pack("<HHII", tag, dtype, nvalues, new_offset)
                b.write(buffer)

                # now write the payload at the outlying position and then come
                # back to the same position in the file stream
                cpos = b.tell()
                b.seek(new_offset)

                format = "<" + DATATYPE2FMT[dtype]["format"] * nvalues
                buffer = struct.pack(format, *payload)
                b.write(buffer)

                # keep track of the next position to write out-of-IFD data
                after_ifd_position = b.tell()
                b.seek(cpos)

            else:

                # the payload DOES fit into the TIFF tag entry
                # write the tag metadata
                buffer = struct.pack("<HHI", tag, dtype, nvalues)
                b.write(buffer)

                payload_format = DATATYPE2FMT[dtype]["format"] * nvalues

                # we may need to alter the output format
                if payload_format in ["H", "B", "I"]:
                    # just write it as an integer
                    payload_format = "I"

                if tag in (34665, 34853):

                    # special case for an EXIF or GPS IFD
                    buffer = struct.pack("<I", after_ifd_position)
                    b.write(buffer)
                    b.seek(after_ifd_position)
                    after_ifd_position = self.write_ifd(b, payload)

                else:

                    buffer = struct.pack("<" + payload_format, *payload)
                    b.write(buffer)

        return after_ifd_position

    def append_xmp_uuid_box(self):
        """Append an XMP UUID box onto the end of the JPEG 2000 file if there
        was an XMP tag in the TIFF IFD.
        """

        if self.xmp_data is None:
            return

        if not self.create_xmp_uuid:
            return

        # create the XMP UUID
        the_uuid = jp2box.UUID("be7acfcb-97a9-42e8-9c71-999491e3afac")
        box_length = len(self.xmp_data) + 8
        uuid_box = jp2box.UUIDBox(the_uuid, self.xmp_data, box_length)
        with self.jp2_path.open(mode="ab") as f:
            uuid_box.write(f)

    def rewrap_for_icc_profile(self):
        """Consume an ICC profile, if one is there."""
        if self.icc_profile is None and self.include_icc_profile:
            self.logger.warning("No ICC profile was found.")

        if self.icc_profile is None or not self.include_icc_profile:
            return

        self.logger.info(
            "Consuming an ICC profile into JP2 color specification box."
        )

        colr = jp2box.ColourSpecificationBox(
            method=RESTRICTED_ICC_PROFILE,
            precedence=0,
            icc_profile=self.icc_profile
        )

        # construct the new set of JP2 boxes, insert the color specification
        # box with the ICC profile
        jp2 = Jp2k(self.jp2_path)
        boxes = jp2.box
        boxes[2].box = [boxes[2].box[0], colr]

        # re-wrap the codestream, involves a file copy
        tmp_filename = str(self.jp2_path) + ".tmp"

        with open(tmp_filename, mode="wb") as tfile:
            jp2.wrap(tfile.name, boxes=boxes)

        shutil.move(tmp_filename, self.jp2_path)



# -*- coding:  utf-8 -*-
"""
Test suite for codestream oddities
"""

# Standard library imports ...
import importlib.resources as ir
from io import BytesIO
import pathlib
import struct
import tempfile
import unittest
import warnings

# Local imports ...
import glymur
from glymur import Jp2k, Jp2kr
from glymur.jp2box import InvalidJp2kError
from . import fixtures


class TestSuite(fixtures.TestCommon):
    """Test suite for codestreams."""

    def setUp(self):
        super().setUp()

        self.p0_03 = ir.files("tests.data").joinpath("p0_03.j2k")
        self.p0_06 = ir.files("tests.data").joinpath("p0_06.j2k")
        self.p1_06 = ir.files("tests.data").joinpath("p1_06.j2k")
        self.issue142 = ir.files("tests.data").joinpath("issue142.j2k")
        self.htj2k = ir.files("tests.data").joinpath("oj-ht-byte.jph")

    def test_cap_marker_segment(self):
        """
        SCENARIO:  the file has a CAP marker segment for the 3rd segment

        EXPECTED RESULT:  the segment metadata is verified
        """
        j = Jp2k(self.htj2k)
        cap = j.codestream.segment[2]

        self.assertEqual(cap.pcap, 131072)
        self.assertEqual(cap.ccap, (3,))

    def test_unrecognized_marker(self):
        """
        SCENARIO:  There is an unrecognized marker just after an SOT marker but
        before the EOC marker.  All markers must have a leading byte value of
        0xff.

        EXPECTED RESULT:  InvalidJp2kError
        """
        with open(self.temp_j2k_filename, mode="wb") as tfile:
            with open(self.j2kfile, "rb") as ifile:
                # Everything up until the SOT marker.
                read_buffer = ifile.read(98)
                tfile.write(read_buffer)

                # Write the bad marker 0xd900
                read_buffer = struct.pack(">H", 0xD900)
                tfile.write(read_buffer)

                # Get the rest of the input file.
                read_buffer = ifile.read()
                tfile.write(read_buffer)
                tfile.flush()

            with self.assertRaises(InvalidJp2kError):
                Jp2k(tfile.name).get_codestream(header_only=False)

    def test_bad_tile_part_pointer(self):
        """
        SCENARIO:  A bad SOT marker segment is encountered (Psot value pointing
        far beyond the end of the EOC marker) when requesting a fully parsed
        codestream.

        EXPECTED RESULT:  InvalidJp2kError
        """
        with open(self.temp_jp2_filename, "wb") as ofile:
            with open(self.jp2file, "rb") as ifile:
                # Copy up until Psot field.
                ofile.write(ifile.read(204))

                # Write a bad Psot value.
                ofile.write(struct.pack(">I", 2000000))

                # copy the rest of the file as-is.
                ifile.seek(208)
                ofile.write(ifile.read())
                ofile.flush()

        j = Jp2kr(self.temp_jp2_filename)
        with self.assertRaises(InvalidJp2kError):
            j.get_codestream(header_only=False)

    def test_tile_height_is_zero(self):
        """
        Scenario:  A tile has height of zero.

        Expected result:  ZeroDivisionError

        Original test file was input/nonregression/2539.pdf.SIGFPE.706.1712.jp2
        """
        fp = BytesIO()

        buffer = struct.pack(">H", 47)  # length

        # kwargs = {'rsiz': 1,
        #           'xysiz': (1000, 1000),
        #           'xyosiz': (0, 0),
        #           'xytsiz': (0, 1000),
        #           'xytosiz': (0, 0),
        #           'Csiz': 3,
        #           'bitdepth': (8, 8, 8),
        #           'signed':  (False, False, False),
        #           'xyrsiz': ((1, 1, 1), (1, 1, 1)),
        #           'length': 47,
        #           'offset': 2}
        buffer += struct.pack(">HIIIIIIIIH", 1, 1000, 1000, 0, 0, 0, 1000, 0, 0, 3)  # noqa : E501
        buffer += struct.pack(">BBBBBBBBB", 7, 1, 1, 7, 1, 1, 7, 1, 1)
        fp.write(buffer)
        fp.seek(0)

        with self.assertRaises(ZeroDivisionError):
            glymur.codestream.Codestream._parse_siz_segment(fp)

    def test_invalid_codestream_past_header(self):
        """
        Scenario:  the codestream is ok thru the header, but invalid after
        that.  The codestream header for the complete test file ends at byte

        Expected result:  InvalidJp2kError
        """
        path = ir.files("tests.data").joinpath("p1_06.j2k")

        with tempfile.TemporaryDirectory() as tdir:
            with open(path, mode="rb") as ifile:
                with open(pathlib.Path(tdir) / "tmp.j2k", mode="wb") as ofile:
                    ofile.write(ifile.read(555))

                with self.assertRaises(InvalidJp2kError):
                    j = Jp2k(pathlib.Path(tdir) / "tmp.j2k")
                    j.get_codestream(header_only=False)

    def test_tlm_segment(self):
        """
        Verify parsing of the TLM segment.

        In this case there's only a single tile.
        """
        path = ir.files("tests.data").joinpath("p0_06.j2k")
        j2k = Jp2k(path)

        buffer = b"\xffU\x00\x08\x00@\x00\x00YW"
        b = BytesIO(buffer[2:])
        segment = j2k.codestream._parse_tlm_segment(b)

        self.assertEqual(segment.ztlm, 0)

        # ttlm is an array, but None is the singleton element
        self.assertIsNone(segment.ttlm.item())

        self.assertEqual(segment.ptlm, (22871,))

    def test_ppt_segment(self):
        """
        Verify parsing of the PPT segment
        """
        path = ir.files("tests.data").joinpath("p1_06.j2k")
        j2k = Jp2k(path)
        c = j2k.get_codestream(header_only=False)
        self.assertEqual(c.segment[6].zppt, 0)

    def test_plt_segment(self):
        """
        Verify parsing of the PLT segment
        """
        path = ir.files("tests.data").joinpath("issue142.j2k")
        c = Jp2k(path).get_codestream(header_only=False)
        self.assertEqual(c.segment[7].zplt, 0)
        self.assertEqual(len(c.segment[7].iplt), 59)

    def test_ppm_segment(self):
        """
        Verify parsing of the PPM segment
        """
        with warnings.catch_warnings():

            # Lots of things wrong with this file.
            warnings.simplefilter("ignore")

            jp2 = Jp2k(ir.files("tests.data").joinpath("edf_c2_1178956.jp2"))

        c = jp2.get_codestream()
        self.assertEqual(c.segment[2].zppm, 0)
        self.assertEqual(len(c.segment[2].data), 9)

    def test_crg_segment(self):
        """
        Verify parsing of the CRG segment
        """
        path = ir.files("tests.data").joinpath("p0_03.j2k")
        j2k = Jp2k(path)
        c = j2k.get_codestream()
        self.assertEqual(c.segment[6].xcrg, (65424,))
        self.assertEqual(c.segment[6].ycrg, (32558,))

    def test_rgn_segment(self):
        """
        Verify parsing of the RGN segment
        """
        path = ir.files("tests.data").joinpath("p0_06.j2k")
        j2k = Jp2k(path)
        c = j2k.get_codestream()
        self.assertEqual(c.segment[-1].crgn, 0)
        self.assertEqual(c.segment[-1].srgn, 0)
        self.assertEqual(c.segment[-1].sprgn, 11)

    def test_reserved_marker_segment(self):
        """
        SCENARIO:  Rewrite a J2K file to include a marker segment with a
        reserved marker 0xff6f (65391).

        EXPECTED RESULT:  The marker segment should be properly parsed.
        """

        with open(self.temp_j2k_filename, "wb") as tfile:
            with open(self.j2kfile, "rb") as ifile:
                # Everything up until the first QCD marker.
                read_buffer = ifile.read(65)
                tfile.write(read_buffer)

                # Write the new marker segment, 0xff6f = 65391
                read_buffer = struct.pack(">HHB", int(65391), int(3), int(0))
                tfile.write(read_buffer)

                # Get the rest of the input file.
                read_buffer = ifile.read()
                tfile.write(read_buffer)
                tfile.flush()

        codestream = Jp2k(tfile.name).get_codestream()

        self.assertEqual(codestream.segment[3].marker_id, "0xff6f")
        self.assertEqual(codestream.segment[3].length, 3)
        self.assertEqual(codestream.segment[3].data, b"\x00")

    def test_siz_segment_ssiz_unsigned(self):
        """ssiz attribute to be removed in future release"""
        j = Jp2k(self.jp2file)
        codestream = j.get_codestream()

        # The ssiz attribute was simply a tuple of raw bytes.
        # The first 7 bits are interpreted as the bitdepth, the MSB determines
        # whether or not it is signed.
        self.assertEqual(codestream.segment[1].ssiz, (7, 7, 7))

    def test_626(self):
        """
        Scenario:  After parsing the SOC and SIZ segments, an unknown segment
        (probably invalid) is hit, and then the file ends, leaving us trying
        to interpret EOF as another marker segment.

        Expected result:  InvalidJp2kError
        """
        path = ir.files("tests.data").joinpath("issue626.j2k")
        with self.assertRaises(InvalidJp2kError):
            Jp2k(path)


class TestCodestreamRepr(unittest.TestCase):

    def setUp(self):
        self.jp2file = glymur.data.nemo()

    def tearDown(self):
        pass

    def test_soc(self):
        """Test SOC segment repr"""
        segment = glymur.codestream.SOCsegment()
        newseg = eval(repr(segment))
        self.assertEqual(newseg.marker_id, "SOC")

    def test_siz(self):
        """Test SIZ segment repr"""
        kwargs = {
            "rsiz": 0,
            "xysiz": (2592, 1456),
            "xyosiz": (0, 0),
            "xytsiz": (2592, 1456),
            "xytosiz": (0, 0),
            "Csiz": 3,
            "bitdepth": (8, 8, 8),
            "signed": (False, False, False),
            "xyrsiz": ((1, 1, 1), (1, 1, 1)),
        }
        segment = glymur.codestream.SIZsegment(**kwargs)
        newseg = eval(repr(segment))
        self.assertEqual(newseg.marker_id, "SIZ")
        self.assertEqual(newseg.xsiz, 2592)
        self.assertEqual(newseg.ysiz, 1456)
        self.assertEqual(newseg.xosiz, 0)
        self.assertEqual(newseg.yosiz, 0)
        self.assertEqual(newseg.xtsiz, 2592)
        self.assertEqual(newseg.ytsiz, 1456)
        self.assertEqual(newseg.xtosiz, 0)
        self.assertEqual(newseg.ytosiz, 0)

        self.assertEqual(newseg.xrsiz, (1, 1, 1))
        self.assertEqual(newseg.yrsiz, (1, 1, 1))
        self.assertEqual(newseg.bitdepth, (8, 8, 8))
        self.assertEqual(newseg.signed, (False, False, False))



# standard library imports
import importlib.resources as ir
import pathlib
import shutil
import struct
import sys
import tempfile
import unittest
from uuid import UUID
import warnings

# 3rd party library imports
import imageio.v3 as iio
import numpy as np
import skimage

# Local imports
import glymur
from glymur import Jp2k, command_line
from glymur.core import SRGB
from . import fixtures
from .fixtures import OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG


@unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
class TestSuite(fixtures.TestCommon):

    @classmethod
    def setUpClass(cls):

        root = ir.files('tests.data.skimage')

        cls.astronaut8 = root.joinpath('astronaut8.tif')
        cls.astronaut_u16 = root.joinpath('astronaut_uint16.tif')
        cls.astronaut_s_u16 = root.joinpath('astronaut_s_uint16.tif')
        cls.astronaut8_stripped = root.joinpath('astronaut8_stripped.tif')
        cls.astronaut_ycbcr_jpeg_tiled = root.joinpath('astronaut_ycbcr_jpeg_tiled.tif')  # noqa : E501
        cls.moon = root.joinpath('moon.tif')
        cls.moon_3x3 = root.joinpath('moon_3x3.tif')
        cls.moon_3stripped = root.joinpath('moon3_stripped.tif')
        cls.moon3_partial_last_strip = root.joinpath('moon3_partial_last_strip.tif')  # noqa : E501
        cls.ycbcr_bg = root.joinpath('ycbcr_bg.tif')
        cls.ycbcr_stripped = root.joinpath('ycbcr_stripped.tif')
        cls.stripped = root.joinpath('stripped.tif')
        cls.moon63 = root.joinpath('moon63.tif')

        test_tiff_dir = tempfile.mkdtemp()
        cls.test_tiff_path = pathlib.Path(test_tiff_dir)

        cls.setup_exif(cls.test_tiff_path / 'exif.tif')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.test_tiff_path)

    @classmethod
    def setup_exif(cls, path):
        """
        Create a simple TIFF file that is constructed to contain an EXIF IFD.
        """

        # main TIFF header @ 0
        # image data @ 8
        # main IFD @ 65544 = 256*256 + 8 (2 + 12*12 = 146 bytes)
        # main IDF data @ 65690  = main_ifd + 2 + 12 * 12 + 4
        #
        # strip offsets @ 65694 (16 bytes)
        # strip byte counts @ 65710 (16 bytes)
        # xmp data @ 65726 (12532 bytes)
        # camera ID data @ 78258 (8 bytes)
        #
        # exif IFD @ 78266 (2 + 2*12 + 4 = 30 bytes)
        # exif IFD data @ 78296 (6 bytes)
        with path.open(mode='wb') as f:

            w = 256
            h = 256
            rps = 64
            header_length = 8

            # write the header (8 bytes).  The IFD will follow the image data
            # (256x256 bytes), so the offset to the IFD will be 8 + h * w.
            main_ifd_offset = header_length + h * w
            buffer = struct.pack('<BBHI', 73, 73, 42, main_ifd_offset)
            f.write(buffer)

            # write the image data, 4 64x256 strips of all zeros
            strip = bytes([0] * rps * w)
            f.write(strip)
            f.write(strip)
            f.write(strip)
            f.write(strip)

            # write an IFD with 12 tags
            main_ifd_data_offset = main_ifd_offset + 2 + 12 * 12 + 4

            buffer = struct.pack('<H', 12)
            f.write(buffer)

            # width and length and bitspersample
            buffer = struct.pack('<HHII', 256, 4, 1, w)
            f.write(buffer)
            buffer = struct.pack('<HHII', 257, 4, 1, h)
            f.write(buffer)
            buffer = struct.pack('<HHII', 258, 4, 1, 8)
            f.write(buffer)

            # photometric
            buffer = struct.pack('<HHII', 262, 4, 1, 1)
            f.write(buffer)

            # strip offsets
            buffer = struct.pack('<HHII', 273, 4, 4, main_ifd_data_offset)
            f.write(buffer)

            # spp
            buffer = struct.pack('<HHII', 277, 4, 1, 1)
            f.write(buffer)

            # rps
            buffer = struct.pack('<HHII', 278, 4, 1, 64)
            f.write(buffer)

            # strip byte counts
            buffer = struct.pack('<HHII', 279, 4, 4, main_ifd_data_offset + 16)
            f.write(buffer)

            # pagenumber
            buffer = struct.pack('<HHIHH', 297, 3, 2, 1, 0)
            f.write(buffer)

            # XMP
            xmp_path = ir.files('tests.data').joinpath('issue555.xmp')
            txt = xmp_path.read_text()
            xmp = txt + '\0'
            buffer = struct.pack(
                '<HHII', 700, 1, len(xmp), main_ifd_data_offset + 32
            )
            f.write(buffer)

            # exif tag
            # write it AFTER lensinfo, which is 8 chars
            exif_ifd_offset = main_ifd_data_offset + 32 + len(xmp) + 8
            buffer = struct.pack('<HHII', 34665, 4, 1, exif_ifd_offset)
            f.write(buffer)

            # lensmodel
            offset = main_ifd_data_offset + 32 + len(xmp)
            buffer = struct.pack('<HHII', 50708, 2, 8, offset)
            f.write(buffer)

            # terminate the IFD
            buffer = struct.pack('<I', 0)
            f.write(buffer)

            # write the strip offsets here
            buffer = struct.pack(
                '<IIII', 8, 8 + rps*w, 8 + 2*rps*w, 8 + 3*rps*w
            )
            f.write(buffer)

            # write the strip byte counts
            buffer = struct.pack('<IIII', rps*w, rps*w, rps*w, rps*w)
            f.write(buffer)

            # write the XMP data
            f.write(xmp.encode('utf-8'))

            # write the camera ID
            f.write("abcdefg\x00".encode('utf-8'))

            # write a minimal Exif IFD
            buffer = struct.pack('<H', 2)
            f.write(buffer)

            # exposure program
            buffer = struct.pack('<HHIHH', 34850, 3, 1, 2, 0)
            f.write(buffer)

            # lens model
            data_location = exif_ifd_offset + 2 + 2*12 + 4
            buffer = struct.pack('<HHII', 42036, 2, 6, data_location)
            f.write(buffer)

            # terminate the IFD
            buffer = struct.pack('<I', 0)
            f.write(buffer)

            data = 'Canon\0'.encode('utf-8')
            buffer = struct.pack('<BBBBBB', *data)
            f.write(buffer)

        cls.exif = path

    def test_psnr_commandline(self):
        """
        SCENARIO:  Convert TIFF file to JP2, specify psnr via the command line

        EXPECTED RESULT:  data matches
        """
        sys.argv = [
            '', str(self.moon), str(self.temp_jp2_filename),
            '--psnr', '30', '35', '40', '0'
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        d = {}
        for layer in range(4):
            j.layer = layer
            d[layer] = j[:]

        truth = iio.imread(self.moon)

        with warnings.catch_warnings():
            # MSE is zero for that first image, resulting in a divide-by-zero
            # warning
            warnings.simplefilter('ignore')
            psnr = [
                skimage.metrics.peak_signal_noise_ratio(truth, d[j])
                for j in range(4)
            ]

        # That first image should be lossless.
        self.assertTrue(np.isinf(psnr[0]))

        # None of the subsequent images should have inf PSNR.
        self.assertTrue(not np.any(np.isinf(psnr[1:])))

        # PSNR should increase for the remaining images.
        self.assertTrue(np.all(np.diff(psnr[1:])) > 0)

    def test_layers_commandline(self):
        """
        SCENARIO:  Convert TIFF file to JP2 with multiple compression layers
        using the command line.

        EXPECTED RESULT:  data matches, number of layers is 3
        """
        sys.argv = [
            '',
            str(self.astronaut_ycbcr_jpeg_tiled), str(self.temp_jp2_filename),
            '--cratio', '200', '50', '10'
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        actual = j[:]
        self.assertEqual(actual.shape, (512, 512, 3))

        c = j.get_codestream()
        self.assertEqual(c.segment[2].layers, 3)

    def test_commandline_tiff2jp2(self):
        """
        Scenario:  patch sys such that we can run the command line tiff2jp2
        script.

        Expected Results:  Same as test_astronaut.
        """
        sys.argv = [
            '', str(self.astronaut8), str(self.temp_jp2_filename),
            '--tilesize', '32', '32'
        ]
        command_line.tiff2jp2()

        jp2 = Jp2k(self.temp_jp2_filename)
        actual = jp2[:]
        expected = skimage.io.imread(self.astronaut8)

        np.testing.assert_array_equal(actual, expected)

        c = jp2.get_codestream()
        self.assertEqual(c.segment[1].xsiz, 64)
        self.assertEqual(c.segment[1].ysiz, 64)
        self.assertEqual(c.segment[1].xtsiz, 32)
        self.assertEqual(c.segment[1].ytsiz, 32)

    def test_commandline_tiff2jp2_exclude_tags_numeric(self):
        """
        Scenario:  patch sys such that we can run the command line tiff2jp2
        script.  Exclude TileByteCounts and TileByteOffsets, but provide those
        tags as numeric values.

        Expected Results:  Same as test_astronaut.
        """
        sys.argv = [
            '', str(self.astronaut8), str(self.temp_jp2_filename),
            '--tilesize', '32', '32',
            '--exclude-tags', '324', '325'
        ]
        command_line.tiff2jp2()

        jp2 = Jp2k(self.temp_jp2_filename)
        tags = jp2.box[-1].data

        self.assertNotIn('TileByteCounts', tags)
        self.assertNotIn('TileOffsets', tags)

    def test_commandline_tiff2jp2_exclude_tags(self):
        """
        Scenario:  patch sys such that we can run the command line tiff2jp2
        script.  Exclude TileByteCounts and TileByteOffsets

        Expected Results:  TileByteCounts and TileOffsets are not in the EXIF
        UUID.
        """
        sys.argv = [
            '', str(self.astronaut8), str(self.temp_jp2_filename),
            '--tilesize', '32', '32',
            '--exclude-tags', 'tilebytecounts', 'tileoffsets'
        ]
        command_line.tiff2jp2()

        jp2 = Jp2k(self.temp_jp2_filename)
        tags = jp2.box[-1].data

        self.assertNotIn('TileByteCounts', tags)
        self.assertNotIn('TileOffsets', tags)

    def test_commandline_capture_display_resolution(self):
        """
        Scenario:  patch sys such that we can run the command
        line tiff2jp2 script.  Supply the --capture-resolution and
        --display-resolution arguments.

        Expected Result:  The last box is a ResolutionBox.
        """
        vresc, hresc = 0.1, 0.2
        vresd, hresd = 0.3, 0.4

        sys.argv = [
            '', str(self.exif), str(self.temp_jp2_filename),
            '--capture-resolution', str(vresc), str(hresc),
            '--display-resolution', str(vresd), str(hresd),
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # the resolution superbox is appended in the jp2 header box.
        # the exit uuid comes later
        self.assertEqual(j.box[-1].box_id, 'uuid')

        self.assertEqual(j.box[2].box[2].box_id, 'res ')

        self.assertEqual(j.box[2].box[2].box[0].box_id, 'resc')
        self.assertEqual(j.box[2].box[2].box[0].vertical_resolution, vresc)
        self.assertEqual(j.box[2].box[2].box[0].horizontal_resolution, hresc)

        self.assertEqual(j.box[2].box[2].box[1].box_id, 'resd')
        self.assertEqual(j.box[2].box[2].box[1].vertical_resolution, vresd)
        self.assertEqual(j.box[2].box[2].box[1].horizontal_resolution, hresd)

    def test_commandline__capture_display_resolution__tilesize(self):
        """
        Scenario:  patch sys such that we can run the command line
        tiff2jp2 script.  Supply the --tilesize, --capture-resolution
        and --display-resolution arguments.

        Expected Result:  The last box is a ResolutionBox.
        """
        vresc, hresc = 0.1, 0.2
        vresd, hresd = 0.3, 0.4

        sys.argv = [
            '', str(self.exif), str(self.temp_jp2_filename),
            '--tilesize', '64', '64',
            '--capture-resolution', str(vresc), str(hresc),
            '--display-resolution', str(vresd), str(hresd),
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # the resolution superbox is appended in the jp2 header box.
        # the exit uuid comes later
        self.assertEqual(j.box[-1].box_id, 'uuid')

        self.assertEqual(j.box[2].box[2].box_id, 'res ')

        self.assertEqual(j.box[2].box[2].box[0].box_id, 'resc')
        self.assertEqual(j.box[2].box[2].box[0].vertical_resolution, vresc)
        self.assertEqual(j.box[2].box[2].box[0].horizontal_resolution, hresc)

        self.assertEqual(j.box[2].box[2].box[1].box_id, 'resd')
        self.assertEqual(j.box[2].box[2].box[1].vertical_resolution, vresd)
        self.assertEqual(j.box[2].box[2].box[1].horizontal_resolution, hresd)

    def test_commandline_tiff2jp2_xmp_uuid(self):
        """
        Scenario:  patch sys such that we can run the command line tiff2jp2
        script.  Use the --create-xmp-uuid option.

        Expected Result:  An Exif UUID is appended to the end of the
        JP2 file, and then an XMP UUID is appended.
        """
        sys.argv = [
            '', str(self.exif), str(self.temp_jp2_filename),
            '--tilesize', '64', '64',
            '--create-xmp-uuid'
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # first we find the Exif UUID, then the XMP UUID.
        actual = j.box[-2].uuid
        expected = UUID(bytes=b'JpgTiffExif->JP2')
        self.assertEqual(actual, expected)

        actual = j.box[-1].uuid
        expected = UUID('be7acfcb-97a9-42e8-9c71-999491e3afac')
        self.assertEqual(actual, expected)
        self.assertEqual(
            j.box[-1].data.getroot().values(), ['Public XMP Toolkit Core 3.5']
        )

    def test_icc_profile_commandline(self):
        """
        Scenario:  The input TIFF has the ICC profile tag.  Provide the
        --include-icc-profile argument.

        Expected Result.  The ICC profile is verified in the
        ColourSpecificationBox.
        """
        path = ir.files('tests.data').joinpath('basn6a08.tif')
        buffer = path.read_bytes()
        ifd = glymur.lib.tiff.tiff_header(buffer)
        icc_profile = bytes(ifd['ICCProfile'])

        sys.argv = [
            '', str(path), str(self.temp_jp2_filename),
            '--include-icc-profile'
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # The colour specification box has the profile
        self.assertEqual(j.box[2].box[1].icc_profile, bytes(icc_profile))

    def test_exclude_icc_profile_commandline(self):
        """
        Scenario:  The input TIFF has the ICC profile tag.  Do not provide the
        --include-icc-profile flag.

        Expected Result.  The ColourSpecificationBox is normal (no ICC
        profile).  The ICC profile tag will be present in the
        JpgTiffExif->JP2 UUID box.
        """
        path = ir.files('tests.data').joinpath('basn6a08.tif')

        sys.argv = ['', str(path), str(self.temp_jp2_filename)]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # The colour specification box does not have the profile
        colr = j.box[2].box[1]
        self.assertEqual(colr.method, glymur.core.ENUMERATED_COLORSPACE)
        self.assertEqual(colr.precedence, 0)
        self.assertEqual(colr.approximation, 0)
        self.assertEqual(colr.colorspace, SRGB)
        self.assertIsNone(colr.icc_profile)

    def test_exclude_icc_profile_commandline__exclude_from_uuid(self):
        """
        Scenario:  The input TIFF has the ICC profile tag.  Do not specify
        the --include-icc-profile flag.  Specify the 34675 (ICCProfile) tag
        in the --exclude-tags flag.

        Expected Result.  The ICC profile is verified to not be present in the
        ColourSpecificationBox.  The ICC profile tag will be not present in the
        JpgTiffExif->JP2 UUID box.
        """
        path = ir.files('tests.data').joinpath('basn6a08.tif')

        sys.argv = [
            '', str(path), str(self.temp_jp2_filename),
            '--exclude-tags', 'ICCProfile',
        ]
        command_line.tiff2jp2()

        j = Jp2k(self.temp_jp2_filename)

        # The colour specification box does not have the profile
        colr = j.box[2].box[1]
        self.assertEqual(colr.method, glymur.core.ENUMERATED_COLORSPACE)
        self.assertEqual(colr.precedence, 0)
        self.assertEqual(colr.approximation, 0)
        self.assertEqual(colr.colorspace, SRGB)
        self.assertIsNone(colr.icc_profile)

        # the exif UUID box does not have the profile
        self.assertNotIn('ICCProfile', j.box[-1].data)



"""Test suite specifically targeting JP2 box layout.
"""

# Standard library imports ...
import importlib.resources as ir
from io import BytesIO
import os
import pathlib
import pickle
import re
import shutil
import struct
import tempfile
from uuid import UUID
import unittest
import warnings

# Third party library imports ...
import lxml.etree as ET
import numpy as np

# Local imports ...
import glymur
from glymur import Jp2k, Jp2kr
from glymur.jp2box import (
    ColourSpecificationBox,
    ContiguousCodestreamBox,
    FileTypeBox,
    ImageHeaderBox,
    JP2HeaderBox,
    JPEG2000SignatureBox,
    BitsPerComponentBox,
    PaletteBox,
    CaptureResolutionBox,
    UnknownBox,
    InvalidJp2kError,
)
from glymur.core import COLOR, OPACITY, SRGB, GREYSCALE
from glymur.core import RED, GREEN, BLUE, GREY, WHOLE_IMAGE
from . import fixtures
from .fixtures import OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG


@unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
class TestDataEntryURL(fixtures.TestCommon):
    """Test suite for DataEntryURL boxes."""

    def test_wrap_greyscale(self):
        """A single component should be wrapped as GREYSCALE."""
        j = Jp2k(self.jp2file)
        data = j[:]
        red = data[:, :, 0]

        # Write it back out as a raw codestream.
        file1 = self.test_dir_path / "file1.j2k"
        j2k = glymur.Jp2k(file1, data=red)

        # Ok, now rewrap it as JP2.  The colorspace should be GREYSCALE.
        file2 = self.test_dir_path / "file2.jp2"
        jp2 = j2k.wrap(file2)

        self.assertEqual(jp2.box[2].box[1].colorspace, glymur.core.GREYSCALE)

    def test_basic_url(self):
        """Just your most basic URL box."""
        # Wrap our j2k file in a JP2 box along with an interior url box.
        jp2 = Jp2k(self.jp2file)

        url = "http://glymur.readthedocs.org"
        deurl = glymur.jp2box.DataEntryURLBox(0, (0, 0, 0), url)
        boxes = [box for box in jp2.box if box.box_id != "uuid"]
        boxes.append(deurl)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            jp22 = jp2.wrap(tfile.name, boxes=boxes)

        actdata = [box.box_id for box in jp22.box]
        expdata = ["jP  ", "ftyp", "jp2h", "jp2c", "url "]
        self.assertEqual(actdata, expdata)
        self.assertEqual(jp22.box[4].version, 0)
        self.assertEqual(jp22.box[4].flag, (0, 0, 0))
        self.assertEqual(jp22.box[4].url, url)

    def test_null_termination(self):
        """I.9.3.2 specifies that location field must be null terminated."""
        jp2 = Jp2k(self.jp2file)

        url = "http://glymur.readthedocs.org"
        deurl = glymur.jp2box.DataEntryURLBox(0, (0, 0, 0), url)
        boxes = [box for box in jp2.box if box.box_id != "uuid"]
        boxes.append(deurl)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            jp22 = jp2.wrap(tfile.name, boxes=boxes)

            self.assertEqual(jp22.box[-1].length, 42)

            # Go to the last box.  Seek past the L, T, version,
            # and flag fields.
            with open(tfile.name, "rb") as fptr:
                fptr.seek(jp22.box[-1].offset + 4 + 4 + 1 + 3)

                nbytes = jp22.box[-1].offset + jp22.box[-1].length - fptr.tell()
                read_buffer = fptr.read(nbytes)
                read_url = read_buffer.decode("utf-8")
                self.assertEqual(url + chr(0), read_url)


@unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
class TestChannelDefinition(fixtures.TestCommon):
    """Test suite for channel definition boxes."""

    @classmethod
    def setUpClass(cls):
        """Need a one_plane plane image for greyscale testing."""
        j2k = Jp2k(glymur.data.goodstuff())
        data = j2k[:]

        cls.planes_dir = pathlib.Path(tempfile.mkdtemp())

        # Write the first component back out to file.
        cls.one_plane = cls.planes_dir / "one_plane.j2k"
        Jp2k(cls.one_plane, data=data[:, :, 0])

        # Write the first two components back out to file.
        cls.two_planes = cls.planes_dir / "two_planes.j2k"
        Jp2k(cls.two_planes, data=data[:, :, 0:2])

        # Write four components back out to file.
        cls.four_planes = cls.planes_dir / "four_planes.j2k"
        shape = (data.shape[0], data.shape[1], 1)
        alpha = np.zeros((shape), dtype=data.dtype)
        data4 = np.concatenate((data, alpha), axis=2)
        Jp2k(cls.four_planes, data=data4)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.planes_dir)

    def setUp(self):
        super().setUp()

        j2k = Jp2k(self.j2kfile)
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)

        self.jp2b = JPEG2000SignatureBox()
        self.ftyp = FileTypeBox()
        self.jp2h = JP2HeaderBox()
        self.jp2c = ContiguousCodestreamBox()
        self.ihdr = ImageHeaderBox(
            height=height, width=width, num_components=num_components
        )
        self.ihdr1 = ImageHeaderBox(height=height, width=width, num_components=1)  # noqa : E501
        self.ihdr2 = ImageHeaderBox(height=height, width=width, num_components=2)  # noqa : E501
        self.ihdr4 = ImageHeaderBox(height=height, width=width, num_components=4)  # noqa : E501
        self.colr_rgb = ColourSpecificationBox(colorspace=SRGB)
        self.colr_gr = ColourSpecificationBox(colorspace=GREYSCALE)

    def test_cdef_no_inputs(self):
        """channel_type and association are required inputs."""
        with self.assertRaises(TypeError):
            glymur.jp2box.ChannelDefinitionBox()

    def test_rgb_with_index(self):
        """Just regular RGB."""
        j2k = Jp2k(self.j2kfile)
        channel_type = [COLOR, COLOR, COLOR]
        association = [RED, GREEN, BLUE]
        cdef = glymur.jp2box.ChannelDefinitionBox(
            index=[0, 1, 2], channel_type=channel_type, association=association
        )
        boxes = [self.ihdr, self.colr_rgb, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)

            jp2 = Jp2k(tfile.name)
            jp2h = jp2.box[2]
            boxes = [box.box_id for box in jp2h.box]
            self.assertEqual(boxes, ["ihdr", "colr", "cdef"])
            self.assertEqual(jp2h.box[2].index, (0, 1, 2))
            self.assertEqual(jp2h.box[2].channel_type, (COLOR, COLOR, COLOR))
            self.assertEqual(jp2h.box[2].association, (RED, GREEN, BLUE))

    def test_rgb(self):
        """Just regular RGB, but don't supply the optional index."""
        j2k = Jp2k(self.j2kfile)
        channel_type = [COLOR, COLOR, COLOR]
        association = [RED, GREEN, BLUE]
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr, self.colr_rgb, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)

            jp2 = Jp2k(tfile.name)
            jp2h = jp2.box[2]
            boxes = [box.box_id for box in jp2h.box]
            self.assertEqual(boxes, ["ihdr", "colr", "cdef"])
            self.assertEqual(jp2h.box[2].index, (0, 1, 2))
            self.assertEqual(jp2h.box[2].channel_type, (COLOR, COLOR, COLOR))
            self.assertEqual(jp2h.box[2].association, (RED, GREEN, BLUE))

    def test_rgba(self):
        """Just regular RGBA."""
        j2k = Jp2k(self.four_planes)
        channel_type = (COLOR, COLOR, COLOR, OPACITY)
        association = (RED, GREEN, BLUE, WHOLE_IMAGE)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr4, self.colr_rgb, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)

            jp2 = Jp2k(tfile.name)

            jp2h = jp2.box[2]
            boxes = [box.box_id for box in jp2h.box]
            self.assertEqual(boxes, ["ihdr", "colr", "cdef"])
            self.assertEqual(jp2h.box[2].index, (0, 1, 2, 3))
            self.assertEqual(jp2h.box[2].channel_type, channel_type)
            self.assertEqual(jp2h.box[2].association, association)

    def test_bad_rgba(self):
        """R, G, and B must be specified."""
        j2k = Jp2k(self.four_planes)
        channel_type = (COLOR, COLOR, OPACITY, OPACITY)
        association = (RED, GREEN, BLUE, WHOLE_IMAGE)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr, self.colr_rgb, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_grey(self):
        """Just regular greyscale."""
        j2k = Jp2k(self.one_plane)
        channel_type = (COLOR,)
        association = (GREY,)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr1, self.colr_gr, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)

            jp2 = Jp2k(tfile.name)
            jp2h = jp2.box[2]
            boxes = [box.box_id for box in jp2h.box]
            self.assertEqual(boxes, ["ihdr", "colr", "cdef"])
            self.assertEqual(jp2h.box[2].index, (0,))
            self.assertEqual(jp2h.box[2].channel_type, channel_type)
            self.assertEqual(jp2h.box[2].association, association)

    def test_grey_alpha(self):
        """Just regular greyscale plus alpha."""
        j2k = Jp2k(self.two_planes)
        channel_type = (COLOR, OPACITY)
        association = (GREY, WHOLE_IMAGE)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr2, self.colr_gr, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)

            jp2 = Jp2k(tfile.name)
            jp2h = jp2.box[2]
            boxes = [box.box_id for box in jp2h.box]
            self.assertEqual(boxes, ["ihdr", "colr", "cdef"])
            self.assertEqual(jp2h.box[2].index, (0, 1))
            self.assertEqual(jp2h.box[2].channel_type, channel_type)
            self.assertEqual(jp2h.box[2].association, association)

    def test_bad_grey_alpha(self):
        """A greyscale image with alpha layer must specify a color channel"""
        j2k = Jp2k(self.two_planes)

        channel_type = (OPACITY, OPACITY)
        association = (GREY, WHOLE_IMAGE)

        # This cdef box
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )
        boxes = [self.ihdr, self.colr_gr, cdef]
        self.jp2h.box = boxes
        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises((OSError, RuntimeError)):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_only_one_cdef_in_jp2h(self):
        """There can only be one channel definition box in the jp2 header."""
        j2k = Jp2k(self.j2kfile)

        channel_type = (COLOR, COLOR, COLOR)
        association = (RED, GREEN, BLUE)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )

        boxes = [self.ihdr, cdef, self.colr_rgb, cdef]
        self.jp2h.box = boxes

        boxes = [self.jp2b, self.ftyp, self.jp2h, self.jp2c]

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_not_in_jp2h(self):
        """need cdef in jp2h"""
        j2k = Jp2k(self.j2kfile)
        boxes = [self.ihdr, self.colr_rgb]
        self.jp2h.box = boxes

        channel_type = (COLOR, COLOR, COLOR)
        association = (RED, GREEN, BLUE)
        cdef = glymur.jp2box.ChannelDefinitionBox(
            channel_type=channel_type, association=association
        )

        boxes = [self.jp2b, self.ftyp, self.jp2h, cdef, self.jp2c]

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises((RuntimeError, OSError)):
                j2k.wrap(tfile.name, boxes=boxes)


class TestFileTypeBox(fixtures.TestCommon):
    """Test suite for ftyp box issues."""

    def test_jph(self):
        """
        SCENARIO:  JPH box

        EXPECTED RESULT:  The brand is verified.
        """
        path = ir.files("tests.data").joinpath("oj-ht-byte.jph")
        j = Jp2kr(path)
        self.assertEqual(j.box[1].brand, "jph ")

    def test_bad_brand_on_parse(self):
        """
        SCENARIO:  The JP2 file file type box does not contain a valid brand.

        EXPECTED RESULT:  RuntimeError
        """
        path = ir.files("tests.data").joinpath("issue396.jp2")
        with warnings.catch_warnings():
            # Lots of things wrong with this file.
            warnings.simplefilter("ignore")
            with self.assertRaises(InvalidJp2kError):
                Jp2k(path)

    def test_brand_unknown(self):
        """A ftyp box brand must be 'jp2 ' or 'jpx '."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            ftyp = glymur.jp2box.FileTypeBox(brand="jp3")
        with tempfile.TemporaryFile() as tfile:
            with self.assertRaises(InvalidJp2kError):
                ftyp.write(tfile)

    def test_cl_entry_unknown(self):
        """A ftyp box cl list can only contain 'jp2 ', 'jpx ', or 'jpxb'."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Bad compatibility list item.
            ftyp = glymur.jp2box.FileTypeBox(compatibility_list=["jp3"])
        with tempfile.TemporaryFile() as tfile:
            with self.assertRaises(InvalidJp2kError):
                ftyp.write(tfile)

    def test_cl_entry_not_utf8(self):
        """A ftyp box cl list entry must be utf-8 decodable."""
        with open(self.jp2file, mode="rb") as f:
            data = f.read()

        # Replace bytes 28-32 with bad utf-8 data
        data = data[:28] + b"\xff\xff\xff\xff" + data[32:]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            tfile.write(data)
            tfile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(tfile.name)


class TestResolutionBoxes(fixtures.TestCommon):
    """
    Test suite for resolution boxes
    """

    def test_invalid_resolution(self):
        """
        Scenario:  the supplied resolution cannot be represented

        Expected Result:  ValueError when trying to write
        """
        resc = CaptureResolutionBox(1e-200, 0.5)
        with self.assertRaises(ValueError):
            with tempfile.NamedTemporaryFile(mode="wb") as tf:
                resc.write(tf)

    def test_repr(self):
        """
        Verify __repr__ method on resolution boxes.
        """
        resc = glymur.jp2box.CaptureResolutionBox(0.5, 2.5)
        resd = glymur.jp2box.DisplayResolutionBox(2.5, 0.5)
        res_super_box = glymur.jp2box.ResolutionBox(box=[resc, resd])

        newbox = eval(repr(res_super_box))

        self.assertEqual(newbox.box_id, "res ")
        self.assertEqual(newbox.box[0].box_id, "resc")
        self.assertEqual(newbox.box[0].vertical_resolution, 0.5)
        self.assertEqual(newbox.box[0].horizontal_resolution, 2.5)
        self.assertEqual(newbox.box[1].box_id, "resd")
        self.assertEqual(newbox.box[1].vertical_resolution, 2.5)
        self.assertEqual(newbox.box[1].horizontal_resolution, 0.5)

    def test_pickling(self):
        """
        SCENARIO:  pickle a ResolutionBox

        Expected Results:  should not error out
        """
        vres = 0.5
        hres = 2.5
        resc = glymur.jp2box.CaptureResolutionBox(vres, hres)
        resd = glymur.jp2box.DisplayResolutionBox(vres, hres)
        rbox = glymur.jp2box.ResolutionBox(box=[resc, resd])

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            rbox.write(tfile)

        with open(self.temp_jp2_filename, mode="rb") as tfile:
            tfile.seek(8)
            rbox_read = glymur.jp2box.ResolutionBox.parse(tfile, 0, 44)

            with tempfile.NamedTemporaryFile(mode="wb") as t2:
                pickle.dump(rbox_read, t2)

    def test_resolution_superbox(self):
        """
        SCENARIO:  write a resolution superbox

        Expected Results:  do not error out, can parse the written box
        """
        vres = 0.5
        hres = 2.5
        resc = glymur.jp2box.CaptureResolutionBox(vres, hres)
        resd = glymur.jp2box.DisplayResolutionBox(vres, hres)
        rbox = glymur.jp2box.ResolutionBox(box=[resc, resd])

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            rbox.write(tfile)

        with open(self.temp_jp2_filename, mode="rb") as tfile:
            tfile.seek(8)
            rbox_read = glymur.jp2box.ResolutionBox.parse(tfile, 0, 44)

        self.assertEqual(rbox_read.box[0].vertical_resolution, vres)
        self.assertEqual(rbox_read.box[0].horizontal_resolution, hres)
        self.assertEqual(rbox_read.box[1].vertical_resolution, vres)
        self.assertEqual(rbox_read.box[1].horizontal_resolution, hres)

    def test_write_capture_resolution_box_high_res(self):
        """
        SCENARIO:  write a capture resolution box with no information other
        than the floating point components.  The components have a very high
        resolution.

        Expected Results:  do not error out, can parse the written box
        """
        vres = 1.8738870547679375e29
        hres = 3333444444.44444
        resc = glymur.jp2box.CaptureResolutionBox(vres, hres)

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            resc.write(tfile)

        with open(self.temp_jp2_filename, mode="rb") as tfile:
            tfile.seek(8)
            resc_read = glymur.jp2box.CaptureResolutionBox.parse(tfile, 8, 18)

        np.testing.assert_allclose(vres, resc_read.vertical_resolution, rtol=1e-6)  # noqa : E501
        np.testing.assert_allclose(hres, resc_read.horizontal_resolution, rtol=1e-6)  # noqa : E501

    def test_write_capture_resolution_box_low_res(self):
        """
        SCENARIO:  write a capture resolution box with no information other
        than the floating point components.  The components have a very low
        resolution.

        Expected Results:  do not error out, can parse the written box
        """
        vres = 1.8738870547679375e-29
        hres = 0.333344444444444
        resc = glymur.jp2box.CaptureResolutionBox(vres, hres)

        with open(self.temp_jp2_filename, mode="wb") as tfile:
            resc.write(tfile)

        with open(self.temp_jp2_filename, mode="rb") as tfile:
            tfile.seek(8)
            resc_read = glymur.jp2box.CaptureResolutionBox.parse(tfile, 8, 18)

        np.testing.assert_allclose(vres, resc_read.vertical_resolution, rtol=1e-6)  # noqa : E501
        np.testing.assert_allclose(hres, resc_read.horizontal_resolution, rtol=1e-6)  # noqa : E501


class TestPaletteBox(fixtures.TestCommon):
    """Test suite for pclr box instantiation."""

    def test_writing_with_different_bitdepths(self):
        """Bitdepths must be the same when writing."""
        palette = np.array([[255, 0, 255], [0, 255, 0]], dtype=np.uint16)
        bps = (8, 16, 8)
        signed = (False, False, False)
        pclr = glymur.jp2box.PaletteBox(palette, bits_per_component=bps, signed=signed)  # noqa : E501
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(InvalidJp2kError):
                pclr.write(tfile)

    def test_signed_components(self):
        """
        Scenario:  a palette box with signed components is read (not supported)

        Expected Results InvalidJp2kError
        """
        b = BytesIO()

        # L, T
        b.write(struct.pack(">I4s", 20, b"pclr"))

        # Palette is 2 rows, 3 columns
        ncols = 3
        nrows = 2
        b.write(struct.pack(">HB", nrows, ncols))

        # Force the leading bit to be 1, which signifies "signed"
        bps = (np.int8(-1), np.int8(-1), np.int8(-1))
        b.write(struct.pack("bbb", *bps))

        # Write the palette itself.
        #
        buffer = np.int8([[0, 0, 0], [127, 127, 127]])
        b.write(struct.pack("BBB", *buffer[0]))
        b.write(struct.pack("BBB", *buffer[1]))

        # Seek back to point after L, T
        b.seek(8)
        with self.assertRaises(InvalidJp2kError):
            PaletteBox.parse(b, 8, 20)


class TestAppend(fixtures.TestCommon):
    """Tests for append method."""

    def test_append_xml(self):
        """Should be able to append an XML box."""
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            shutil.copyfile(self.jp2file, tfile.name)

            jp2 = Jp2k(tfile.name)
            b = BytesIO(b'<?xml version="1.0"?><data>0</data>')
            doc = ET.parse(b)
            xmlbox = glymur.jp2box.XMLBox(xml=doc)
            jp2.append(xmlbox)

            # The sequence of box IDs should be the same as before, but with an
            # xml box at the end.
            box_ids = [box.box_id for box in jp2.box]
            expected = ["jP  ", "ftyp", "jp2h", "jp2c", "xml "]
            self.assertEqual(box_ids, expected)
            self.assertEqual(ET.tostring(jp2.box[-1].xml.getroot()), b"<data>0</data>")  # noqa : E501

    def test_only_jp2_allowed_to_append(self):
        """Only JP2 files are allowed to be appended."""
        with open(self.temp_j2k_filename, mode="wb") as tfile:
            shutil.copyfile(self.j2kfile, tfile.name)

            j2k = Jp2k(tfile.name)

            # Make an XML box.  XML boxes should always be appendable to jp2
            # files.
            the_xml = ET.fromstring('<?xml version="1.0"?><data>0</data>')
            xmlbox = glymur.jp2box.XMLBox(xml=the_xml)
            with self.assertRaises(RuntimeError):
                j2k.append(xmlbox)

    def test_length_field_is_zero(self):
        """L=0 (length field in box header) is handled.

        L=0 implies that the containing box is the last box.  If this is not
        handled properly, the appended box is never seen.
        """
        baseline_jp2 = Jp2k(self.jp2file)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with open(self.jp2file, "rb") as ifile:
                # Everything up until the jp2c box.
                offset = baseline_jp2.box[-1].offset
                tfile.write(ifile.read(offset))

                # Write the L, T fields of the jp2c box such that L == 0
                write_buffer = struct.pack(">I4s", int(0), b"jp2c")
                tfile.write(write_buffer)

                # Write out the rest of the codestream.
                ifile.seek(offset + 8)
                tfile.write(ifile.read())
                tfile.flush()

            jp2 = Jp2k(tfile.name)
            b = BytesIO(b'<?xml version="1.0"?><data>0</data>')
            doc = ET.parse(b)
            xmlbox = glymur.jp2box.XMLBox(xml=doc)
            jp2.append(xmlbox)

            # The sequence of box IDs should be the same as before, but with an
            # xml box at the end.
            box_ids = [box.box_id for box in jp2.box]
            expected = ["jP  ", "ftyp", "jp2h", "jp2c", "xml "]
            self.assertEqual(box_ids, expected)
            self.assertEqual(ET.tostring(jp2.box[-1].xml.getroot()), b"<data>0</data>")  # noqa : E501

    def test_append_allowable_boxes(self):
        """Only XML boxes are allowed to be appended."""
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            shutil.copyfile(self.jp2file, tfile.name)

            jp2 = Jp2k(tfile.name)

            # Make a UUID box.  Only XMP UUID boxes can currently be appended.
            uuid_instance = UUID("00000000-0000-0000-0000-000000000000")
            data = b"0123456789"
            uuidbox = glymur.jp2box.UUIDBox(uuid_instance, data)
            with self.assertRaises(RuntimeError):
                jp2.append(uuidbox)


class TestWrap(fixtures.TestCommon):
    """Tests for wrap method."""

    def verify_wrapped_raw(self, jp2file):
        """Shared fixture"""
        jp2 = Jp2k(jp2file)
        self.assertEqual(len(jp2.box), 4)

        self.assertEqual(jp2.box[0].box_id, "jP  ")
        self.assertEqual(jp2.box[0].offset, 0)
        self.assertEqual(jp2.box[0].length, 12)
        self.assertEqual(jp2.box[0].longname, "JPEG 2000 Signature")

        self.assertEqual(jp2.box[1].box_id, "ftyp")
        self.assertEqual(jp2.box[1].offset, 12)
        self.assertEqual(jp2.box[1].length, 20)
        self.assertEqual(jp2.box[1].longname, "File Type")

        self.assertEqual(jp2.box[2].box_id, "jp2h")
        self.assertEqual(jp2.box[2].offset, 32)
        self.assertEqual(jp2.box[2].length, 45)
        self.assertEqual(jp2.box[2].longname, "JP2 Header")

        self.assertEqual(jp2.box[3].box_id, "jp2c")
        self.assertEqual(jp2.box[3].offset, 77)
        self.assertEqual(jp2.box[3].length, 115228)

        # jp2h super box
        self.assertEqual(len(jp2.box[2].box), 2)

        self.assertEqual(jp2.box[2].box[0].box_id, "ihdr")
        self.assertEqual(jp2.box[2].box[0].offset, 40)
        self.assertEqual(jp2.box[2].box[0].length, 22)
        self.assertEqual(jp2.box[2].box[0].longname, "Image Header")
        self.assertEqual(jp2.box[2].box[0].height, 800)
        self.assertEqual(jp2.box[2].box[0].width, 480)
        self.assertEqual(jp2.box[2].box[0].num_components, 3)
        self.assertEqual(jp2.box[2].box[0].bits_per_component, 8)
        self.assertEqual(jp2.box[2].box[0].signed, False)
        self.assertEqual(jp2.box[2].box[0].compression, 7)
        self.assertEqual(jp2.box[2].box[0].colorspace_unknown, False)
        self.assertEqual(jp2.box[2].box[0].ip_provided, False)

        self.assertEqual(jp2.box[2].box[1].box_id, "colr")
        self.assertEqual(jp2.box[2].box[1].offset, 62)
        self.assertEqual(jp2.box[2].box[1].length, 15)
        self.assertEqual(jp2.box[2].box[1].longname, "Colour Specification")
        self.assertEqual(jp2.box[2].box[1].precedence, 0)
        self.assertEqual(jp2.box[2].box[1].approximation, 0)
        self.assertEqual(jp2.box[2].box[1].colorspace, glymur.core.SRGB)
        self.assertIsNone(jp2.box[2].box[1].icc_profile)

    def test_wrap(self):
        """basic test for rewrapping a j2c file, no specified boxes"""
        j2k = Jp2k(self.j2kfile)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name)
            self.verify_wrapped_raw(tfile.name)

    def test_jpx_to_jp2(self):
        """basic test for rewrapping a jpx file"""
        jpx = Jp2k(self.jpxfile)
        # Use only the signature, file type, header, and 1st codestream.
        lst = [0, 1, 2, 5]
        boxes = [jpx.box[idx] for idx in lst]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            jp2 = jpx.wrap(tfile.name, boxes=boxes)

        # Verify the outer boxes.
        boxes = [box.box_id for box in jp2.box]
        self.assertEqual(boxes, ["jP  ", "ftyp", "jp2h", "jp2c"])

        # Verify the inside boxes.
        boxes = [box.box_id for box in jp2.box[2].box]
        self.assertEqual(boxes, ["ihdr", "colr", "pclr", "cmap"])

        expected_offsets = [0, 12, 40, 887]
        for j, offset in enumerate(expected_offsets):
            self.assertEqual(jp2.box[j].offset, offset)

    def test_wrap_jp2(self):
        """basic test for rewrapping a jp2 file, no specified boxes"""
        j2k = Jp2k(self.jp2file)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            jp2 = j2k.wrap(tfile.name)
        boxes = [box.box_id for box in jp2.box]
        self.assertEqual(boxes, ["jP  ", "ftyp", "jp2h", "jp2c"])

    def test_wrap_jp2_Lzero(self):
        """Wrap jp2 with jp2c box length is zero"""
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with open(self.jp2file, "rb") as ifile:
                tfile.write(ifile.read())
            # Rewrite with codestream length as zero.
            tfile.seek(77)
            tfile.write(struct.pack(">I", 0))
            tfile.flush()

        jp = Jp2k(tfile.name)

        file2 = self.test_dir_path / "file2.jp2"
        jp2 = jp.wrap(file2)
        boxes = [box for box in jp2.box]
        self.assertEqual(boxes[3].length, 1132296)

    def test_wrap_jp2_Lone(self):
        """Wrap jp2 with jp2c box length is 1, implies Q field"""
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with open(self.jp2file, "rb") as ifile:
                tfile.write(ifile.read(77))
                # Write new L, T, Q fields
                tfile.write(struct.pack(">I4sQ", 1, b"jp2c", 1132296 + 8))
                # skip over the old L, T fields
                ifile.seek(85)
                tfile.write(ifile.read())
            tfile.flush()
            jp2 = Jp2k(tfile.name)

        file2 = self.test_dir_path / "file2.jp2"
        jp22 = jp2.wrap(file2)
        self.assertEqual(jp22.box[3].length, 1132296 + 8)

    def test_wrap_compatibility_not_jp2(self):
        """File type compatibility must contain jp2"""
        jp2 = Jp2k(self.jp2file)
        boxes = [box for box in jp2.box]
        boxes[1].compatibility_list = ["jpx "]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                jp2.wrap(tfile.name, boxes=boxes)

    def test_empty_jp2h(self):
        """JP2H box list cannot be empty."""
        jp2 = Jp2k(self.jp2file)
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            boxes = jp2.box
            # Right here the jp2h superbox has two child boxes.  Empty out that
            # list to trigger the error.
            boxes[2].box = []
            with self.assertRaises(RuntimeError):
                jp2.wrap(tfile.name, boxes=boxes)

    def test_default_layout_with_boxes(self):
        """basic test for rewrapping a jp2 file, boxes specified"""
        j2k = Jp2k(self.j2kfile)
        boxes = [
            JPEG2000SignatureBox(),
            FileTypeBox(),
            JP2HeaderBox(),
            ContiguousCodestreamBox(),
        ]
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)
        boxes[2].box = [
            ImageHeaderBox(height=height, width=width, num_components=num_components),  # noqa : E501
            ColourSpecificationBox(colorspace=glymur.core.SRGB),
        ]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            j2k.wrap(tfile.name, boxes=boxes)
            self.verify_wrapped_raw(tfile.name)

    def test_ihdr_not_first_in_jp2h(self):
        """The specification says that ihdr must be the first box in jp2h."""
        j2k = Jp2k(self.j2kfile)
        boxes = [
            JPEG2000SignatureBox(),
            FileTypeBox(),
            JP2HeaderBox(),
            ContiguousCodestreamBox(),
        ]
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)
        boxes[2].box = [
            ColourSpecificationBox(colorspace=glymur.core.SRGB),
            ImageHeaderBox(height=height, width=width, num_components=num_components),  # noqa : E501
        ]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_first_boxes_jp_and_ftyp(self):
        """first two boxes must be jP followed by ftyp"""
        j2k = Jp2k(self.j2kfile)
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)

        jp2b = JPEG2000SignatureBox()
        ftyp = FileTypeBox()
        jp2h = JP2HeaderBox()
        jp2c = ContiguousCodestreamBox()
        colr = ColourSpecificationBox(colorspace=glymur.core.SRGB)
        ihdr = ImageHeaderBox(height=height, width=width, num_components=num_components)  # noqa : E501
        jp2h.box = [ihdr, colr]
        boxes = [ftyp, jp2b, jp2h, jp2c]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_pclr_not_in_jp2h(self):
        """A palette box must reside in a JP2 header box."""
        palette = np.array([[255, 0, 255], [0, 255, 0]], dtype=np.int32)
        bps = (8, 8, 8)
        pclr = glymur.jp2box.PaletteBox(
            palette=palette, bits_per_component=bps, signed=(True, False, True)
        )

        j2k = Jp2k(self.j2kfile)
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)

        jp2b = JPEG2000SignatureBox()
        ftyp = FileTypeBox()
        jp2h = JP2HeaderBox()
        jp2c = ContiguousCodestreamBox()
        colr = ColourSpecificationBox(colorspace=glymur.core.SRGB)
        ihdr = ImageHeaderBox(height=height, width=width, num_components=num_components)  # noqa : E501
        jp2h.box = [ihdr, colr]
        boxes = [jp2b, ftyp, jp2h, jp2c, pclr]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_jp2h_not_preceeding_jp2c(self):
        """jp2h must precede jp2c"""
        j2k = Jp2k(self.j2kfile)
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)

        jp2b = JPEG2000SignatureBox()
        ftyp = FileTypeBox()
        jp2h = JP2HeaderBox()
        jp2c = ContiguousCodestreamBox()
        colr = ColourSpecificationBox(colorspace=glymur.core.SRGB)
        ihdr = ImageHeaderBox(height=height, width=width, num_components=num_components)  # noqa : E501
        jp2h.box = [ihdr, colr]
        boxes = [jp2b, ftyp, jp2c, jp2h]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_missing_codestream(self):
        """Need a codestream box in order to call wrap method."""
        j2k = Jp2k(self.j2kfile)
        codestream = j2k.get_codestream()
        height = codestream.segment[1].ysiz
        width = codestream.segment[1].xsiz
        num_components = len(codestream.segment[1].xrsiz)

        jp2k = JPEG2000SignatureBox()
        ftyp = FileTypeBox()
        jp2h = JP2HeaderBox()
        ihdr = ImageHeaderBox(height=height, width=width, num_components=num_components)  # noqa : E501
        jp2h.box = [ihdr]
        boxes = [jp2k, ftyp, jp2h]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with self.assertRaises(RuntimeError):
                j2k.wrap(tfile.name, boxes=boxes)

    def test_wrap_jpx_to_jp2_with_unadorned_jpch(self):
        """A JPX file rewrapped with plain jpch is not allowed."""
        with open(self.temp_jp2_filename, mode="wb") as tfile1:
            jpx = Jp2k(self.jpxfile)
            boxes = [
                jpx.box[0],
                jpx.box[1],
                jpx.box[2],
                glymur.jp2box.ContiguousCodestreamBox(),
            ]
            with self.assertRaises(RuntimeError):
                jpx.wrap(tfile1.name, boxes=boxes)

    def test_wrap_jpx_to_jp2_with_incorrect_jp2c_offset(self):
        """Reject A JPX file rewrapped with bad jp2c offset."""
        with open(self.temp_jp2_filename, mode="wb") as tfile1:
            jpx = Jp2k(self.jpxfile)
            jpch = jpx.box[5]

            # The offset should be 902.
            jpch.offset = 901
            jpch.length = 313274
            boxes = [jpx.box[0], jpx.box[1], jpx.box[2], jpch]
            with self.assertRaises(RuntimeError):
                jpx.wrap(tfile1.name, boxes=boxes)

    def test_wrap_jpx_to_jp2_with_correctly_specified_jp2c(self):
        """Accept A JPX file rewrapped with good jp2c."""
        with open(self.temp_jp2_filename, mode="wb") as tfile1:
            jpx = Jp2k(self.jpxfile)
            jpch = jpx.box[5]

            # This time get it right.
            jpch.offset = 903
            jpch.length = 313274
            boxes = [jpx.box[0], jpx.box[1], jpx.box[2], jpch]
            jp2 = jpx.wrap(tfile1.name, boxes=boxes)

        act_ids = [box.box_id for box in jp2.box]
        exp_ids = ["jP  ", "ftyp", "jp2h", "jp2c"]
        self.assertEqual(act_ids, exp_ids)

        act_offsets = [box.offset for box in jp2.box]
        exp_offsets = [0, 12, 40, 887]
        self.assertEqual(act_offsets, exp_offsets)

        act_lengths = [box.length for box in jp2.box]
        exp_lengths = [12, 28, 847, 313274]
        self.assertEqual(act_lengths, exp_lengths)

    def test_full_blown_jpx(self):
        """Rewrap a jpx file."""
        with open(self.temp_jp2_filename, mode="wb") as tfile1:
            jpx = Jp2k(self.jpxfile)
            idx = (list(range(5)) + list(range(9, 12)) + list(range(6, 9))) + [12]  # noqa : E501
            boxes = [jpx.box[j] for j in idx]
            jpx2 = jpx.wrap(tfile1.name, boxes=boxes)
            exp_ids = [box.box_id for box in boxes]
            lengths = [box.length for box in jpx.box]
            exp_lengths = [lengths[j] for j in idx]
        act_ids = [box.box_id for box in jpx2.box]
        act_lengths = [box.length for box in jpx2.box]
        self.assertEqual(exp_ids, act_ids)
        self.assertEqual(exp_lengths, act_lengths)


class TestJp2Boxes(fixtures.TestCommon):
    """Tests for canonical JP2 boxes."""

    def test_issue588(self):
        """
        Scenario:  Construct a raw codestream box without parsing it.
        Retrieve the codestream attribute.

        Expected results:  no errors
        """
        j = Jp2k(self.jp2file)

        box = ContiguousCodestreamBox(
            main_header_offset=j.box[-1].main_header_offset,
            length=j.box[-1].length
        )
        box._filename = str(self.jp2file)
        box.codestream

    def test_no_jp2h_box(self):
        """
        SCENARIO:  The JP2/JP2H box is missing

        EXPECTED RESULT:  InvalidJp2kError
        """
        # Write a new JP2 file that omits the IHDR box.
        j = Jp2k(self.jp2file)
        jp2h = [box for box in j.box if box.box_id == "jp2h"][0]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            numbytes = jp2h.offset
            with open(self.jp2file, "rb") as ifile:
                # Write all the way up to the ihdr box
                tfile.write(ifile.read(numbytes))

                # Seek past the ihdr box
                ifile.seek(jp2h.length, os.SEEK_CUR)

                # Write the rest of the JP2 file
                tfile.write(ifile.read(numbytes))

            tfile.flush()

            with self.assertRaises(InvalidJp2kError):
                with warnings.catch_warnings():
                    # Lots of things wrong with this file.
                    warnings.simplefilter("ignore")
                    Jp2k(tfile.name)

    def test_no_ihdr_box(self):
        """
        SCENARIO:  The JP2/IHDR box cannot be parsed.

        EXPECTED RESULT:  InvalidJp2kError
        """
        # Write a new JP2 file that omits the IHDR box.
        j = Jp2k(self.jp2file)
        jp2h = [box for box in j.box if box.box_id == "jp2h"][0]
        ihdr = jp2h.box[0]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            numbytes = ihdr.offset
            with open(self.jp2file, "rb") as ifile:
                # Write all the way up to the ihdr box
                tfile.write(ifile.read(numbytes))

                # Seek past the ihdr box
                ifile.seek(ihdr.length, os.SEEK_CUR)

                # Write the rest of the JP2 file
                tfile.write(ifile.read(numbytes))

            tfile.flush()

            with self.assertRaises(InvalidJp2kError):
                with warnings.catch_warnings():
                    # Lots of things wrong with this file.
                    warnings.simplefilter("ignore")
                    Jp2k(tfile.name)

    def test_no_jp2c_box(self):
        """
        SCENARIO:  The JP2 file has no JP2C box.

        EXPECTED RESULT:  An InvalidJp2kError is issued when the file is
        parsed.
        """
        testfile = ir.files("tests.data").joinpath("no_jp2c.jp2")

        with warnings.catch_warnings():
            # Lots of things wrong with this file.
            warnings.simplefilter("ignore")
            with self.assertRaises(InvalidJp2kError):
                Jp2k(testfile)

    def test_two_jp2c_boxes(self):
        """
        SCENARIO:  There are two jp2c boxes.

        EXPECTED RESULT:  Technically this is valid but it is still possible
        to read the first image.  A warning is issued since the 2nd image
        cannot be accessed.
        """
        # Write a new JP2 file that duplicates the JP2C box.
        j = Jp2k(self.jp2file)
        jp2c = [box for box in j.box if box.box_id == "jp2c"][0]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with open(self.jp2file, "rb") as ifile:
                tfile.write(ifile.read())

                # now tack the jp2c box on again
                ifile.seek(jp2c.offset)
                tfile.write(ifile.read())

            tfile.flush()

            with self.assertWarns(UserWarning):
                j2 = Jp2k(tfile.name)

        np.testing.assert_array_equal(j[:], j2[:])

    def test_two_jp2h_boxes(self):
        """
        SCENARIO:  There are two jp2h boxes and two jp2c boxes.

        EXPECTED RESULT:  Technically this is invalid due to the multiple jp2h
        boxes.  It is still possible to read the first image.  Multiple
        warnings are issued.
        """
        # Write a new JP2 file that duplicates the JP2C box.
        j = Jp2k(self.jp2file)
        jp2h = [box for box in j.box if box.box_id == "jp2h"][0]
        with open(self.temp_jp2_filename, mode="wb") as tfile:
            with open(self.jp2file, "rb") as ifile:
                tfile.write(ifile.read())

                # now tack the jp2c box on again
                ifile.seek(jp2h.offset)
                tfile.write(ifile.read())

            tfile.flush()

            # Warns because of two codestream, warns because of two jp2h boxes.
            with self.assertWarns(UserWarning) as cm:
                j2 = Jp2k(tfile.name)
                self.assertEqual(len(cm.warnings), 2)

        with warnings.catch_warnings():
            # Another warning when this file is read.  It is a library warning
            # so ignore it.
            warnings.simplefilter("ignore")
            np.testing.assert_array_equal(j[:], j2[:])

    def test_default_jp2k(self):
        """Should be able to instantiate a JPEG2000SignatureBox"""
        jp2k = glymur.jp2box.JPEG2000SignatureBox()
        self.assertEqual(jp2k.signature, (13, 10, 135, 10))

    def test_default_ftyp(self):
        """Should be able to instantiate a FileTypeBox"""
        ftyp = glymur.jp2box.FileTypeBox()
        self.assertEqual(ftyp.brand, "jp2 ")
        self.assertEqual(ftyp.minor_version, 0)
        self.assertEqual(ftyp.compatibility_list, ["jp2 "])

    def test_default_ihdr(self):
        """Should be able to instantiate an image header box."""
        ihdr = glymur.jp2box.ImageHeaderBox(
            height=512, width=256, num_components=3
        )
        self.assertEqual(ihdr.height, 512)
        self.assertEqual(ihdr.width, 256)
        self.assertEqual(ihdr.num_components, 3)
        self.assertEqual(ihdr.bits_per_component, 8)
        self.assertFalse(ihdr.signed)
        self.assertFalse(ihdr.colorspace_unknown)

    def test_default_jp2headerbox(self):
        """Should be able to set jp2h boxes."""
        box = JP2HeaderBox()
        box.box = [
            ImageHeaderBox(height=512, width=256),
            ColourSpecificationBox(colorspace=glymur.core.GREYSCALE),
        ]
        self.assertTrue(True)

    def test_default_ccodestreambox(self):
        """Raw instantiation should not produce a main_header."""
        box = ContiguousCodestreamBox()
        self.assertEqual(box.box_id, "jp2c")
        self.assertIsNone(box.codestream)

    def test_codestream_main_header_offset(self):
        """
        main_header_offset is an attribute of the ContiguousCodesStream box
        """
        j = Jp2k(self.jpxfile)
        self.assertEqual(j.box[5].main_header_offset, j.box[5].offset + 8)


class TestRepr(fixtures.TestCommon):
    """Tests for __repr__ methods."""

    def _verify_filetype_box(self, actual, expected):
        """
        All JP2 files should have a brand reading 'jp2 ' and just a single
        entry in the compatibility list, also 'jp2 '.  JPX files can have more
        compatibility items.
        """
        self.assertEqual(actual.brand, expected.brand)
        self.assertEqual(actual.minor_version, expected.minor_version)
        self.assertEqual(actual.minor_version, 0)
        for cl in expected.compatibility_list:
            self.assertIn(cl, actual.compatibility_list)

    def test_default_jp2k(self):
        """Should be able to eval a JPEG2000SignatureBox"""
        jp2k = glymur.jp2box.JPEG2000SignatureBox()

        # Test the representation instantiation.
        newbox = eval(repr(jp2k))
        self.assertTrue(isinstance(newbox, glymur.jp2box.JPEG2000SignatureBox))
        self.assertEqual(newbox.signature, (13, 10, 135, 10))

    def test_unknown(self):
        """Should be able to instantiate an unknown box"""
        box = UnknownBox("bpcc")

        # Test the representation instantiation.
        newbox = eval(repr(box))
        self.assertTrue(isinstance(newbox, glymur.jp2box.UnknownBox))

    def test_bpcc(self):
        """Should be able to instantiate a bpcc box"""
        bpc = (5, 5, 5, 1)
        signed = (False, False, True, False)
        box = BitsPerComponentBox(bpc, signed, length=12, offset=62)

        # Test the representation instantiation.
        newbox = eval(repr(box))
        self.assertEqual(bpc, newbox.bpc)
        self.assertEqual(signed, newbox.signed)

    def test_free(self):
        """Should be able to instantiate a free box"""
        free = glymur.jp2box.FreeBox()

        # Test the representation instantiation.
        newbox = eval(repr(free))
        self.assertTrue(isinstance(newbox, glymur.jp2box.FreeBox))

    def test_nlst(self):
        """Should be able to instantiate a number list box"""
        assn = (0, 1, 2)
        nlst = glymur.jp2box.NumberListBox(assn)

        # Test the representation instantiation.
        newbox = eval(repr(nlst))
        self.assertTrue(isinstance(newbox, glymur.jp2box.NumberListBox))
        self.assertEqual(newbox.associations, (0, 1, 2))

    def test_ftbl(self):
        """Should be able to instantiate a fragment table box"""
        flst = glymur.jp2box.FragmentListBox([89], [1132288], [0])
        ftbl = glymur.jp2box.FragmentTableBox([flst])

        # Test the representation instantiation.
        newbox = eval(repr(ftbl))
        self.assertTrue(isinstance(newbox, glymur.jp2box.FragmentTableBox))

    def test_dref(self):
        """Should be able to instantiate a data reference box"""
        dref = glymur.jp2box.DataReferenceBox()

        # Test the representation instantiation.
        newbox = eval(repr(dref))
        self.assertTrue(isinstance(newbox, glymur.jp2box.DataReferenceBox))

    def test_flst(self):
        """Should be able to instantiate a fragment list box"""
        flst = glymur.jp2box.FragmentListBox([89], [1132288], [0])

        # Test the representation instantiation.
        newbox = eval(repr(flst))
        self.assertTrue(isinstance(newbox, glymur.jp2box.FragmentListBox))
        self.assertEqual(newbox.fragment_offset, [89])
        self.assertEqual(newbox.fragment_length, [1132288])
        self.assertEqual(newbox.data_reference, [0])

    def test_default_cgrp(self):
        """Should be able to instantiate a color group box"""
        cgrp = glymur.jp2box.ColourGroupBox()

        # Test the representation instantiation.
        newbox = eval(repr(cgrp))
        self.assertTrue(isinstance(newbox, glymur.jp2box.ColourGroupBox))

    def test_default_ftyp(self):
        """Should be able to instantiate a FileTypeBox"""
        ftyp = glymur.jp2box.FileTypeBox()

        # Test the representation instantiation.
        newbox = eval(repr(ftyp))
        self._verify_filetype_box(newbox, FileTypeBox())

    def test_colourspecification_box(self):
        """Verify __repr__ method on colr box."""
        # TODO:  add icc_profile
        box = ColourSpecificationBox(colorspace=glymur.core.SRGB)

        newbox = eval(repr(box))
        self.assertEqual(newbox.method, glymur.core.ENUMERATED_COLORSPACE)
        self.assertEqual(newbox.precedence, 0)
        self.assertEqual(newbox.approximation, 0)
        self.assertEqual(newbox.colorspace, glymur.core.SRGB)
        self.assertIsNone(newbox.icc_profile)

    def test_channeldefinition_box(self):
        """Verify __repr__ method on cdef box."""
        channel_type = [COLOR, COLOR, COLOR]
        association = [RED, GREEN, BLUE]
        cdef = glymur.jp2box.ChannelDefinitionBox(
            index=[0, 1, 2], channel_type=channel_type, association=association
        )
        newbox = eval(repr(cdef))
        self.assertEqual(newbox.index, (0, 1, 2))
        self.assertEqual(newbox.channel_type, (COLOR, COLOR, COLOR))
        self.assertEqual(newbox.association, (RED, GREEN, BLUE))

    def test_jp2header_box(self):
        """Verify __repr__ method on ihdr box."""
        ihdr = ImageHeaderBox(100, 200, num_components=3)
        colr = ColourSpecificationBox(colorspace=glymur.core.SRGB)
        jp2h = JP2HeaderBox(box=[ihdr, colr])
        newbox = eval(repr(jp2h))
        self.assertEqual(newbox.box_id, "jp2h")
        self.assertEqual(newbox.box[0].box_id, "ihdr")
        self.assertEqual(newbox.box[1].box_id, "colr")

    def test_imageheader_box(self):
        """Verify __repr__ method on jhdr box."""
        ihdr = ImageHeaderBox(100, 200, num_components=3)

        newbox = eval(repr(ihdr))
        self.assertEqual(newbox.height, 100)
        self.assertEqual(newbox.width, 200)
        self.assertEqual(newbox.num_components, 3)
        self.assertFalse(newbox.signed)
        self.assertEqual(newbox.bits_per_component, 8)
        self.assertEqual(newbox.compression, 7)
        self.assertFalse(newbox.colorspace_unknown)
        self.assertFalse(newbox.ip_provided)

    def test_association_box(self):
        """Verify __repr__ method on asoc box."""
        asoc = glymur.jp2box.AssociationBox()
        newbox = eval(repr(asoc))
        self.assertEqual(newbox.box_id, "asoc")
        self.assertEqual(len(newbox.box), 0)

    def test_codestreamheader_box(self):
        """Verify __repr__ method on jpch box."""
        jpch = glymur.jp2box.CodestreamHeaderBox()
        newbox = eval(repr(jpch))
        self.assertEqual(newbox.box_id, "jpch")
        self.assertEqual(len(newbox.box), 0)

    def test_compositinglayerheader_box(self):
        """Verify __repr__ method on jplh box."""
        jplh = glymur.jp2box.CompositingLayerHeaderBox()
        newbox = eval(repr(jplh))
        self.assertEqual(newbox.box_id, "jplh")
        self.assertEqual(len(newbox.box), 0)

    def test_componentmapping_box(self):
        """Verify __repr__ method on cmap box."""
        cmap = glymur.jp2box.ComponentMappingBox(
            component_index=(0, 0, 0),
            mapping_type=(1, 1, 1),
            palette_index=(0, 1, 2)
        )
        newbox = eval(repr(cmap))
        self.assertEqual(newbox.box_id, "cmap")
        self.assertEqual(newbox.component_index, (0, 0, 0))
        self.assertEqual(newbox.mapping_type, (1, 1, 1))
        self.assertEqual(newbox.palette_index, (0, 1, 2))

    def test_label_box(self):
        """Verify __repr__ method on label box."""
        lbl = glymur.jp2box.LabelBox("this is a test")
        newbox = eval(repr(lbl))
        self.assertEqual(newbox.box_id, "lbl ")
        self.assertEqual(newbox.label, "this is a test")

    def test_data_entry_url_box(self):
        """Verify __repr__ method on data entry url box."""
        version = 0
        flag = (0, 0, 0)
        url = "http://readthedocs.glymur.org"
        box = glymur.jp2box.DataEntryURLBox(version, flag, url)
        newbox = eval(repr(box))
        self.assertEqual(newbox.box_id, "url ")
        self.assertEqual(newbox.version, version)
        self.assertEqual(newbox.flag, flag)
        self.assertEqual(newbox.url, url)

    def test_uuidinfo_box(self):
        """Verify __repr__ method on uinf box."""
        uinf = glymur.jp2box.UUIDInfoBox()
        newbox = eval(repr(uinf))
        self.assertEqual(newbox.box_id, "uinf")
        self.assertEqual(len(newbox.box), 0)

    def test_uuidlist_box(self):
        """Verify __repr__ method on ulst box."""
        uuid1 = UUID("00000000-0000-0000-0000-000000000001")
        uuid2 = UUID("00000000-0000-0000-0000-000000000002")
        uuids = [uuid1, uuid2]
        ulst = glymur.jp2box.UUIDListBox(ulst=uuids)
        newbox = eval(repr(ulst))
        self.assertEqual(newbox.box_id, "ulst")
        self.assertEqual(newbox.ulst[0], uuid1)
        self.assertEqual(newbox.ulst[1], uuid2)

    def test_palette_box(self):
        """Verify Palette box repr."""
        palette = np.array([[255, 0, 1000], [0, 255, 0]], dtype=np.int32)
        bps = (8, 8, 16)
        box = glymur.jp2box.PaletteBox(
            palette=palette, bits_per_component=bps, signed=(True, False, True)
        )

        # Test will fail unless addition imports from numpy are done.
        from numpy import array, int32  # noqa: F401

        newbox = eval(repr(box))
        np.testing.assert_array_equal(newbox.palette, palette)
        self.assertEqual(newbox.bits_per_component, (8, 8, 16))
        self.assertEqual(newbox.signed, (True, False, True))

    def test_xml_box(self):
        """Verify xml box repr."""
        elt = ET.fromstring('<?xml version="1.0"?><data>0</data>')
        tree = ET.ElementTree(elt)
        box = glymur.jp2box.XMLBox(xml=tree)

        pattern = r"""
            glymur.jp2box.XMLBox
            \(xml=<lxml.etree._ElementTree\sobject\sat\s0x[a-fA-F0-9]+>\)
        """
        regex = re.compile(pattern, re.VERBOSE)
        self.assertRegex(repr(box), regex)

    def test_readerrequirements_box(self):
        """Verify rreq repr method."""
        box = glymur.jp2box.ReaderRequirementsBox(
            fuam=160,
            dcm=192,
            standard_flag=(5, 61, 43),
            standard_mask=(128, 96, 64),
            vendor_feature=[],
            vendor_mask=[],
        )
        newbox = eval(repr(box))
        self.assertEqual(box.fuam, newbox.fuam)
        self.assertEqual(box.dcm, newbox.dcm)
        self.assertEqual(box.standard_flag, newbox.standard_flag)
        self.assertEqual(box.standard_mask, newbox.standard_mask)
        self.assertEqual(box.vendor_feature, newbox.vendor_feature)
        self.assertEqual(box.vendor_mask, newbox.vendor_mask)

    def test_uuid_box_generic(self):
        """Verify uuid repr method."""
        uuid_instance = UUID("00000000-0000-0000-0000-000000000000")
        data = b"0123456789"
        box = glymur.jp2box.UUIDBox(the_uuid=uuid_instance, raw_data=data)

        # Since the raw_data parameter is a sequence of bytes which could be
        # quite long, don't bother trying to make it conform to eval(repr()).
        pattern = r"""
            glymur.jp2box.UUIDBox\(
                UUID\('00000000-0000-0000-0000-000000000000'\),\s
                raw_data=<byte\sarray\s10\selements>
            \)
        """
        regex = re.compile(pattern, re.VERBOSE)
        self.assertRegex(repr(box), regex)

    def test_xmp_uuid_box_repr(self):
        """
        Scenario:  Run an XMP UUID box thru repr.

        Expected Result:  validated against a regex.

        """
        the_uuid = UUID("be7acfcb-97a9-42e8-9c71-999491e3afac")
        raw_data = (
            ir.files("tests.data")
            .joinpath("simple_rdf.txt")
            .read_text()
            .encode("utf-8")
        )

        box = glymur.jp2box.UUIDBox(the_uuid=the_uuid, raw_data=raw_data)

        pattern = r"""
            glymur.jp2box.UUIDBox\(
                UUID\('be7acfcb-97a9-42e8-9c71-999491e3afac'\),\s
                raw_data=<byte\sarray\s410\selements>
            \)
        """
        regex = re.compile(pattern, re.VERBOSE)
        self.assertRegex(repr(box), regex)

    def test_contiguous_codestream_box(self):
        """Verify contiguous codestream box repr method."""
        jp2file = glymur.data.nemo()
        jp2 = Jp2k(jp2file)
        box = jp2.box[-1]

        # Difficult to eval(repr()) this, so just match the general pattern.
        pattern = r"""
            glymur.jp2box.ContiguousCodeStreamBox\(
                codestream=<glymur.codestream.Codestream\sobject\s
                at\s0x([a-fA-F0-9]*)>
            \)
        """
        regex = re.compile(pattern, re.VERBOSE)
        self.assertRegex(repr(box), regex)



"""These tests are for edge cases where OPENJPEG does not exist, but
OPENJP2 may be present in some form or other.
"""

# Standard library imports ...
import contextlib
import importlib
import os
import pathlib
import platform
import unittest
from unittest.mock import patch
import warnings

# Local imports ...
import glymur
from glymur import Jp2k
from .fixtures import TestCommon, OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG  # noqa : E501


@contextlib.contextmanager
def chdir(dirname=None):
    """
    This context manager restores the value of the current working directory
    (cwd) after the enclosed code block completes or raises an exception.  If a
    directory name is supplied to the context manager then the cwd is changed
    prior to running the code block.

    Shamelessly lifted from
    http://www.astropython.org/snippet/2009/10/chdir-context-manager
    """
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


@patch("glymur.config.glymurrc_fname", lambda: None)
class TestSuitePathToLibrary(TestCommon):
    """
    Test the path determined for the openjp2 library.

    This test suite assumes NO rc config file, so we have to force that code
    path to not run in case we are actively using it.  This should not be a
    problem in CI environments, just development environments.
    """

    def setUp(self):
        super().setUp()

    def tearDown(self):
        """
        Do the normal tear-down, but then make sure that we reload the openjp2
        library normally.  Otherwise a bad library configuration might
        contaminate the next test.
        """
        super().tearDown()

        importlib.reload(glymur)
        importlib.reload(glymur.lib.openjp2)

    @patch("glymur.config.find_library")
    @patch("glymur.config.platform.system")
    def test_via_ctypes(self, mock_platform_system, mock_find_library):
        """
        SCENARIO:  the platform is not anaconda and not MacPorts.  The ctypes
        module finds the library.

        EXPECTED RESULT:  the path of the openjp2 library is on standard
        system paths
        """
        mock_platform_system.return_value = "not darwin"
        mock_find_library.return_value = "/usr/lib/libopenjp2.so"

        actual = glymur.config._determine_full_path("openjp2")
        expected = pathlib.Path("/usr/lib/libopenjp2.so")

        self.assertEqual(actual, expected)

    @patch("glymur.config.read_config_file")
    @patch("glymur.config.pathlib")
    @patch("glymur.config.platform.system")
    def test_get_library_on_cygwin(
        self, mock_platform_system, mock_pathlib, mock_read_config_file
    ):
        """
        SCENARIO:  the platform is cygwin and there is no config file

        EXPECTED RESULT:  return the path of the openjpeg library on cygwin
        """
        mock_platform_system.return_value = "CYGWIN"
        mock_pathlib.Path.return_value.glob.return_value = [
            pathlib.Path("/usr/bin/cygopenjp2.dll")
        ]
        mock_read_config_file.return_value = None

        actual = glymur.config._determine_full_path("openjp2")
        expected = pathlib.Path("/usr/bin/cygopenjp2.dll")
        self.assertEqual(actual, expected)

    @patch("glymur.config.read_config_file")
    @patch("glymur.config.pathlib")
    @patch("glymur.config.platform.system")
    def test_get_library_on_cygwin_when_it_does_not_exist(
        self, mock_platform_system, mock_pathlib, mock_read_config_file
    ):
        """
        SCENARIO:  the platform is cygwin, there is no config file, and openjp2
        is not installed

        EXPECTED RESULT:  None
        """
        mock_platform_system.return_value = "CYGWIN"
        mock_pathlib.Path.return_value.glob.return_value = []
        mock_read_config_file.return_value = None

        actual = glymur.config._determine_full_path("openjp2")

        self.assertIsNone(actual)


@patch("glymur.config.glymurrc_fname", lambda: None)
class TestSuite(TestCommon):
    """
    This test suite assumes NO rc config file, so we have to force that code
    path to not run in case we are actively using it.  This should not be a
    problem in CI environments, just development environments.
    """

    @patch("glymur.config._determine_full_path")
    @patch("glymur.config.platform.system")
    def test_library_cannot_be_loaded(self, mock_platform_system, mock_determine_path):  # noqa : E501
        """
        SCENARIO:  the platform is Linux but the ctypes module does not find
        libopenjp2

        EXPECTED RESULT:  glymur_config returns None and issues a warning
        """
        mock_platform_system.return_value = "Linux"
        mock_determine_path.return_value = "/does/not/exist"

        with self.assertWarns(UserWarning):
            actual = glymur.config.glymur_config("openjp2")
        self.assertIsNone(actual)

    @patch("glymur.config.find_library")
    @patch("glymur.config.platform.system")
    def test_tiff_not_via_ctypes(self, mock_platform_system, mock_find_library):
        """
        SCENARIO:  the platform is not anaconda and not MacPorts.  The ctypes
        module does NOT find tiff.

        EXPECTED RESULT:  the path of the tiff library is None
        """
        mock_platform_system.return_value = "not darwin"
        mock_find_library.return_value = None

        actual = glymur.config.glymur_config("tiff")
        self.assertIsNone(actual)

    @patch("glymur.config.find_library")
    @patch("glymur.config.platform.system")
    def test_openjp2_not_via_ctypes(self, mock_platform_system, mock_find_library):  # noqa : E501
        """
        SCENARIO:  the platform is not anaconda and not MacPorts.  The ctypes
        module does NOT find openjp2.

        EXPECTED RESULT:  the path of the openjp2 library is None
        """
        mock_platform_system.return_value = "not darwin"
        mock_find_library.return_value = None

        actual = glymur.config.glymur_config("openjp2")
        self.assertIsNone(actual)

    @unittest.skipIf(platform.system() == "Windows", "nonsensical on windows")
    @patch("glymur.config.platform.system")
    @patch("pathlib.Path.home")
    def test_config_dir_on_windows(
        self,
        mock_pathlib_path_home,
        mock_platform_system
    ):
        """
        SCENARIO:  the XDG_CONFIG_HOME environment variable is not present, the
        os.name *IS* 'nt'.  Don't bother running on windows because that's what
        we are trying to test from other platforms.

        EXPECTED RESULT:  the path to the configuration directory should be
        under the home directory
        """
        mock_platform_system.return_value = "Windows"

        expected_path = pathlib.Path("/neither/here/nor/there")
        mock_pathlib_path_home.return_value = expected_path

        with patch.dict("os.environ", values=()):
            # Just make sure XDG_CONFIG_HOME is not present.
            actual = glymur.config.get_configdir()
        self.assertEqual(actual, expected_path / "glymur")


class TestSuiteOptions(unittest.TestCase):

    def setUp(self):
        glymur.reset_option("all")

    def tearDown(self):
        glymur.reset_option("all")

    def test_reset_single_option(self):
        """
        Verify a single option can be reset.
        """
        glymur.set_option("print.codestream", True)
        glymur.reset_option("print.codestream")
        self.assertTrue(glymur.get_option("print.codestream"))

    def test_bad_reset(self):
        """
        Verify exception when a bad option is given to reset
        """
        with self.assertRaises(KeyError):
            glymur.reset_option("blah")

    def test_bad_deprecated_print_option(self):
        """
        Verify exception when a bad option is given to old set_printoption
        """
        with self.assertRaises(KeyError):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                glymur.set_printoptions(blah="value-blah")


@unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
class TestSuiteConfigFile(TestCommon):
    """Test suite for configuration file operation."""

    def setUp(self):
        super().setUp()

        # Setup a config root for glymur.
        self.config_root = self.test_dir_path / "config"
        self.config_root.mkdir()

        self.glymur_configdir = self.config_root / "glymur"
        self.glymur_configdir.mkdir()

        self.config_file = self.glymur_configdir / "glymurrc"

    def tearDown(self):
        """
        Do the normal tear-down, but then make sure that we reload the openjp2
        library normally.  Otherwise a bad library configuration might
        contaminate the next test.
        """
        super().tearDown()

        importlib.reload(glymur)
        importlib.reload(glymur.lib.openjp2)

    def test_config_file_via_environ(self):
        """
        SCENARIO:  Specify the configuration file via an environment variable.

        EXPECTED RESULT:  The openjp2 library is loaded normally.
        """
        with self.config_file.open("wt") as f:
            f.write("[library]\n")

            # Need to reliably recover the location of the openjp2 library,
            # so using '_name' appears to be the only way to do it.
            libloc = glymur.lib.openjp2.OPENJP2._name
            line = "openjp2: {0}\n".format(libloc)
            f.write(line)

        new = {"XDG_CONFIG_HOME": str(self.config_root)}
        with patch.dict("os.environ", new):
            importlib.reload(glymur.lib.openjp2)
            Jp2k(self.jp2file)

    def test_config_file_without_library_section(self):
        """
        SCENARIO:  A config directory is specified via the environment
        variable, but the config file does not have a library section.

        EXPECTED RESULT:  Just don't error out.
        """
        with self.config_file.open("wt") as f:
            f.write("[testing]\n")
            f.write("opj_data_root: blah\n")

        new = {"XDG_CONFIG_HOME": str(self.config_root)}
        with patch.dict("os.environ", new):
            importlib.reload(glymur.lib.openjp2)
            # It's enough that we did not error out
            self.assertTrue(True)

    def test_config_dir_but_no_config_file(self):
        """
        SCENARIO:  A config directory is specified via the environment
        variable, but no config file exists.

        EXPECTED RESULT:  openjp2 library is still loaded from default
        location, but a warning is also issued.
        """

        new = {"XDG_CONFIG_HOME": str(self.config_root)}
        with patch.dict("os.environ", new):
            importlib.reload(glymur.lib.openjp2)
            self.assertIsNotNone(glymur.lib.openjp2.OPENJP2)

    def test_config_file_in_current_directory(self):
        """
        SCENARIO:  A configuration file exists in the current directory.

        EXPECTED RESULT:  the path to the specified openjp2 library is returned
        """
        new_lib_dir = self.test_dir_path / "lib"
        new_lib_dir.mkdir()

        expected = new_lib_dir / "libopenjp2.dylib"

        with self.config_file.open("wt") as f:
            f.write("[library]\n")
            f.write(f"openjp2: {expected}\n")

        with chdir(self.glymur_configdir):
            # Should be able to load openjp2 as before.
            actual = glymur.config.read_config_file("openjp2")

        self.assertEqual(actual, expected)
